import re
import tkinter as tk
import webbrowser
from enum import Enum
from tkinter import ttk


class MdField(tk.Text):
    class FS(Enum):
        H1 = 'A'
        H2 = 'B'
        H3 = 'C'
        ITALIC = 'I'
        BOLD = 'D'
        LIST_U = 'U'
        LIST_O = 'O'
        LINK = 'L'
        END = 'X'
        FINALEND = 'E'

        @classmethod
        def all_styles(cls):
            result = []
            for s in cls:
                result.append(s.__str__())
            return result

        def __str__(self):
            return '#' + self.value

    def __init__(self, frame: ttk.Widget, font=("Arial", 12)):
        super().__init__(frame, wrap="word", font=font)
        self._font = font

        self._tags = {
            "A": {'font': (None, self._font[1] + 6, "bold")},
            "B": {'font': (None, self._font[1] + 4, "bold")},
            "C": {'font': (None, self._font[1] + 2, "bold")},
            "D": {'font': (None, None, "bold")},
            "I": {'font': (None, None, "italic")},
            "O": {'font': (None, None, None)},
            "U": {'font': (None, None, None)},
            "L": {'font': (None, None, None),'foreground': 'blue', 'underline': True},
            "E": {'font': (None, None, None)},
        }

    def _prepare_tags(self, tags):
        tags = [t.value for t in tags]
        tag = ''.join(tags)
        if tag == '' or tag in self.tag_names():
            return
        font = [self._font[0], self._font[1], []]
        rest_style_info = {}
        for t in tags:
            for key, val in self._tags[t].items():
                if key != 'font':
                    rest_style_info[key] = val

            for i in range(2):
                if self._tags[t]['font'][i] is not None:
                    font[i] = self._tags[t]['font'][i]
            if self._tags[t]['font'][2] is not None:
                font[2].append(self._tags[t]['font'][2])

        self.tag_config(tag, font=(font[0], font[1], ' '.join(font[2])), **rest_style_info)

        return tag

    def _change_cursor(self, event, cursor_type):
        self.config(cursor=cursor_type)
        self.current_cursor = cursor_type  # Store cursor state

    def set_md_file(self, md_fp):
        with open(md_fp, "r", encoding="utf-8") as f:
            self.set_md_text(f.read())

    def open_link(self, url):
        webbrowser.open(url)

    def set_md_text(self, md_text):

        self.tag_configure("hyperlink", foreground="blue", underline=True)
        self.tag_bind("hyperlink", "<Enter>", lambda e: self._change_cursor(e, "hand2"))
        self.tag_bind("hyperlink", "<Leave>", lambda e: self._change_cursor(e, "xterm"))
        

        # Convert headers
        md_text = re.sub(r"^# (.*)", fr"{self.FS.H1}\1{self.FS.END}", md_text, flags=re.MULTILINE)
        md_text = re.sub(r"^## (.*)", fr"{self.FS.H2}\1{self.FS.END}", md_text, flags=re.MULTILINE)
        md_text = re.sub(r"^### (.*)", fr"{self.FS.H3}\1{self.FS.END}", md_text, flags=re.MULTILINE)

        # Convert bold (**text** or __text__)
        md_text = re.sub(r"\*\*(.*?)\*\*", fr"{self.FS.BOLD}\1{self.FS.END}", md_text)
        md_text = re.sub(r"__(.*?)__", fr"{self.FS.BOLD}\1{self.FS.END}", md_text)

        # Convert italic (*text* or _text_)
        md_text = re.sub(r"\*(.*?)\*", fr"{self.FS.ITALIC}\1{self.FS.END}", md_text)
        md_text = re.sub(r"_(.*?)_", fr"{self.FS.ITALIC}\1{self.FS.END}", md_text)

        # Convert unordered lists (- item)
        md_text = re.sub(r"^- (.*)", fr"   - {self.FS.LIST_U}\1{self.FS.END}", md_text, flags=re.MULTILINE)
        md_text = re.sub(r"^(\d+)\. (.*)", fr"   \1. {self.FS.LIST_O}\2{self.FS.END}", md_text, flags=re.MULTILINE)
        link_re = re.compile(r"\[(.*?)]\((.*?)\)")
        md_text = re.sub(link_re, fr"{self.FS.LINK}[\1](\2){self.FS.END}", md_text)

        md_text += str(self.FS.FINALEND)
        text = ''



        current_style_tags = []
        last = False
        for i in range(len(md_text) - 1):
            if md_text[i:i + 2] in self.FS.all_styles():
                last = True
                fs = self.FS(md_text[i + 1])
                st = self._prepare_tags(current_style_tags)
                if len(current_style_tags) > 0 and current_style_tags[-1] == self.FS.LINK:
                    res = link_re.findall(text)[0]
                    text = res[0]
                    st = (st, 'hyperlink', res[1])
                    self.tag_configure(res[1])
                    self.tag_bind(res[1], "<Button-1>", lambda x: self.open_link(res[1]))  # Click event

                self.insert("end", text, st)

                text = ''
                if fs == self.FS.END:
                    current_style_tags.pop()
                elif fs != self.FS.FINALEND:
                    current_style_tags.append(fs)
            elif not last:
                text += md_text[i]
            else:
                last = False
        return md_text



import importlib.util
import os
import runpy
import shutil
import tempfile
import uuid
from functools import wraps
from pathlib import Path
from subprocess import Popen, PIPE
from typing import LiteralString, Literal

from ChemDoE.config import ConfigManager
from ChemDoE.registery.result_validater import validate_results
from ChemDoE.registery.loader import read_results
from _pytest.monkeypatch import MonkeyPatch
import sys

def _get_absolute_path(module_name):
    spec = importlib.util.find_spec(module_name)
    if spec and spec.origin:
        return os.path.abspath(spec.origin)
    return None


def _generate_test_args(tmp_path: Path, input_type: Literal['json', 'csv'],
                        output_type: Literal['json', 'csv']):
    # Create temporary input and output files
    input_file_name = f"{input_type}_sample.{input_type}"
    output_file_name = f"out_{output_type}_sample.{output_type}"
    input_file = tmp_path / input_file_name
    output_file = tmp_path / output_file_name

    src = Path(__file__).parent.parent / 'examples'

    shutil.copyfile(src / input_file_name, input_file)


    # Simulate command-line arguments
    return  [sys.argv[0], str(input_file), str(output_file)]


def check_doe_script(input_type: Literal['json', 'csv'], output_type: Literal['json', 'csv']):
    """Pytest decorator to test if the decorated function works according to the requirements of a DoE run script."""

    def decorator(test_func):
        @wraps(test_func)
        def wrapper(*test_args, **test_kwargs):
            with tempfile.TemporaryDirectory() as tmpdir:
                mp = MonkeyPatch()
                new_args = _generate_test_args(Path(tmpdir), input_type, output_type)
                # Simulate command-line arguments
                mp.setattr(sys, "argv", new_args)
                try:
                    test_func(*test_args, **test_kwargs)
                    _load_and_test_results(sys.argv[-1])# Run the test
                finally:
                    mp.undo()  # Restore original sys.argv after the test

        return wrapper

    return decorator


def _load_and_test_results(result_path: str | Path | LiteralString):
    res = read_results(result_path)
    assert validate_results(res) == []



def _from_abs_path_run_python_doe_script(script_path):
    cmd = [sys.executable, script_path, sys.argv[1], sys.argv[2]]
    p = Popen(cmd, stdout=PIPE,
              stderr=PIPE,
              text=True,  # Ensures output is in string format instead of bytes
              bufsize=1,  # Line-buffered output
              universal_newlines=True,  # Handles newlines properly across platforms
              cwd=os.path.dirname(script_path)
              )
    for line in iter(p.stdout.readline, ''):
        print(line, end="")
        
def _from_import_path_run_python_doe_script(script_path):
    runpy.run_module(script_path, run_name="__main__")
    
def register_python_doe_script(script_path, input_type: Literal['json', 'csv'],  output_type: Literal['json', 'csv']):
    """
    Executes a Python script with test DoE Table and adds it to the run scripts of ChemDoE if the results are in the correct format

    :param script_path: Ether the file path of the Python script or the import path
    :param input_type: File format of the input file
    :param output_type: File format of the output file
    :return:
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        sys.argv = _generate_test_args(Path(tmpdir), input_type, output_type)

        if os.path.exists(script_path):
            _from_abs_path_run_python_doe_script(script_path)
            abs_path = os.path.abspath(script_path)
        else:
            abs_path = _get_absolute_path(script_path)
            _from_import_path_run_python_doe_script(script_path)

        _load_and_test_results(sys.argv[-1])
        scripts = ConfigManager().load_scripts()
        values = next((x for x in scripts if x["file"] == abs_path), None)

        if values is None:
            values = {'id': uuid.uuid4().__str__()}
            scripts.append(values)

        values["name"] = abs_path
        values["file_type"] = "Python"
        values["input"] = input_type.upper()
        values["output"] = output_type.upper()
        values["file"] = abs_path
        values["interpreter"] = sys.executable

        ConfigManager().save_scripts(scripts)


def register_r_doe_script(script_path, input_type: Literal['json', 'csv'], output_type: Literal['json', 'csv']):
    """
    Executes an R script with test DoE Table and adds it to the run scripts of ChemDoE if the results are in the correct format

    :param script_path: File path of the R script
    :param input_type: File format of the input file
    :param output_type: File format of the output file
    :return:
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        sys_argv = _generate_test_args(Path(tmpdir), input_type, output_type)
        r = ConfigManager().r_interpreters[0]
        cmd = [r, script_path, sys_argv[1], sys_argv[2]]
        p = Popen(cmd, stdout=PIPE,
                  stderr=PIPE,
                  text=True,  # Ensures output is in string format instead of bytes
                  bufsize=1,  # Line-buffered output
                  universal_newlines=True,  # Handles newlines properly across platforms
                  cwd=os.path.dirname(script_path)
                  )
        for line in iter(p.stdout.readline, ''):
            print(line)
        _load_and_test_results(sys_argv[2])
        scripts = ConfigManager().load_scripts()

        abs_path = os.path.abspath(script_path)
        values = next((x for x in scripts if x["file"] == abs_path), None)

        if values is None:
            values = {'id': uuid.uuid4().__str__()}
            scripts.append(values)

        values["name"] = script_path
        values["file_type"] = "R"
        values["input"] = input_type.upper()
        values["output"] = output_type.upper()
        values["file"] = abs_path
        values["interpreter"] = sys.executable

        ConfigManager().save_scripts(scripts)



import threading
from tkinter import ttk
import tkinter as tk
from typing import Optional, Literal

from chemotion_api import Reaction, Sample
from chemotion_api.collection import Collection, RootCollection

from ChemDoE.icons import IconManager
from ChemDoE.utils.pages import ToolBarPage
from ChemDoE.utils.keyboard_shortcuts import add_placeholder


class ElementTreePage(ToolBarPage):
    number_of_elements_per_page = 10
    search_placeholder = "Enter search text..."

    class TreeElement:
        def __init__(self, title: str):
            self.title = title
            self.tree_node: Optional[str] = None

    class TreeElementElement(TreeElement):
        def __init__(self, rea: Sample | Reaction):
            if isinstance(rea, Sample):
                super().__init__(f"{rea.short_label}: {rea.molecule['cano_smiles']}")
            else:
                super().__init__(f"{rea.short_label}: {rea.name}")

            self.rea = rea


    class TreeElementCollection(TreeElement):
        def __init__(self, col: Collection | RootCollection):
            super().__init__(col.label)
            self.loaded = False
            self.col = col
            self.children: list[ElementTreePage.TreeElementCollection] = []
            self.elements: list[ElementTreePage.TreeElementElement] = []

    def __init__(self, element_type: Literal['Reaction', 'Sample'] = 'Reaction'):
        super().__init__()
        self._tree_structor = ElementTreePage.TreeElementCollection(self.instance.get_root_collection())
        self._tree_structor.tree_node = ""
        self._search_var = tk.StringVar()
        self._collection_registry: dict[str, ElementTreePage.TreeElementCollection] = {}
        self._search_task_id = None
        self._element_type = element_type


    def render(self, container: ttk.Frame):
        super().render(container)
        entry = ttk.Entry(container, textvariable=self._search_var)
        entry.pack(fill="x", padx=5, pady=2)
        entry.bind("<KeyRelease>", lambda e: self._search_tree())
        add_placeholder(entry, self.search_placeholder)

        self.paned_window = ttk.PanedWindow(container, orient=tk.HORIZONTAL)
        self.paned_window.pack(fill=tk.BOTH, expand=True)

        left_frame = ttk.Frame(self.paned_window, width=200, relief=tk.SUNKEN)
        self.paned_window.add(left_frame, weight=3)  # weight allows resizing
        self._render_collection_tree(left_frame)

    def _render_collection_tree(self, container: ttk.Frame):
        self.collection_tree = ttk.Treeview(container)

        self.collection_tree.heading("#0", text=f"Collections & {self._element_type}", anchor="w")
        self._tree_structor.children = self._load_tree(self._tree_structor.col)
        self._fill_tree(self._tree_structor)

        self.collection_tree.bind("<<TreeviewOpen>>", self._on_open)

        self.collection_tree.bind("<ButtonRelease-1>", self._on_element_click)

        self.collection_tree.pack(expand=True, fill="both")

    def set_style(self, style: ttk.Style):
        super().set_style(style)
        style.configure("Treeview.Heading", font=("Arial", 12, "bold"), background="#ffffff", foreground="#000000",
                        borderwidth=0, relief="flat")
        style.map("Treeview.Heading", background=[("active", "#ffffff")], foreground=[("active", "#000000")])
        style.configure("Reaction.Treeview", rowheight=30)

    def _search_tree(self):
        if self._search_task_id is not None:
            self.page_manager.root.after_cancel(self._search_task_id)
        self._search_task_id = self.page_manager.root.after(500, self._filter_tree)

    def _filter_tree(self, root = None, disable_detach:bool = False):
        """Filter treeview based on search input."""
        if root is None:
            root = self._tree_structor
        if isinstance(root, ElementTreePage.TreeElementElement):
            return False
        query = self._search_var.get().lower()
        if len(query) < 2:
            disable_detach = True
        res = False
        for item in root.children + root.elements:
            match = query in item.title.lower()
            if disable_detach or match or self._filter_tree(item):
                if disable_detach or match:
                    self._filter_tree(item, True)
                res = True
                try:# Check if query matches any column
                    self.collection_tree.reattach(item.tree_node, root.tree_node, "end")  #
                except tk.TclError:
                    pass
                if not disable_detach and isinstance(item, ElementTreePage.TreeElementCollection):
                    self.collection_tree.item(item.tree_node, open=True)  #
                    self._load_on_open(item.tree_node)
            else:
                self.collection_tree.detach(item.tree_node)
        return res

    def _load_tree(self, root_col):
        root = []
        for col in root_col.children:
            te = self.TreeElementCollection(col)
            self._collection_registry[str(col.id)] = te
            root.append(te)
            te.children = self._load_tree(col)
        return root

    def _fill_tree(self, root_col: TreeElementCollection, parent="", filter_label=None):
        ret = []
        for te in root_col.children:
            te.tree_node = self.collection_tree.insert(parent, "end", text=te.title, open=False,
                                                  image=IconManager().FOLDER_ICON,
                                                  values=(str(te.col.id), 'Collection'))
            self._fill_tree(te, te.tree_node, filter_label)

            self.collection_tree.insert(te.tree_node, "end", text=f'Create new {self._element_type}', open=False,
                                        image=IconManager().PLUS_ICON,
                                        values=(str(te.col.id), 'NewElement'))
            self._load_on_open(te.tree_node)

        return ret


    def _load_next_page(self, te: TreeElementCollection):
        try:
            if self._element_type == 'Reaction':
                elements = te.col.get_reactions(self.number_of_elements_per_page)
                icon = IconManager().REACTION_ICON
            else:
                elements = te.col.get_samples(self.number_of_elements_per_page)
                icon = IconManager().SAMPLE_ICON
            current_length = len(te.elements)
            all_elements = len(elements)
            if all_elements == current_length:
                return False

            if not self.is_visible:
                return
            self.page_manager.root.after(0, lambda te: self.collection_tree.item(te.tree_node, text=te.title + " ⏳"), te)

            end_length = min(all_elements, current_length + self.number_of_elements_per_page)
            for i in range(current_length, end_length):
                if not self.is_visible:
                    return
                re = ElementTreePage.TreeElementElement(elements[i])
                te.elements.append(re)
                def add_node(re, te):
                    if self.is_visible:
                        re.tree_node = self.collection_tree.insert(te.tree_node, "end", text=re.title, open=False,
                                                image=icon,
                                                values=(str(re.rea.id), "Element"))
                self.page_manager.root.after(0, add_node, re, te)

            if self.is_visible:
                self.page_manager.root.after(0, lambda te: self.collection_tree.item(te.tree_node, text=te.title ), te)
            if len(te.elements) < all_elements:
                self.page_manager.root.after(0, lambda te: self.collection_tree.insert(te.tree_node, "end", text="...Load more...", open=False, image=icon, values=(str(te.col.id), "LOAD")), te)


        except tk.TclError:
            return


    def _on_open(self, event):
        """Triggered when a tree item is opened (expanded)."""

        item = self.collection_tree.focus()  # Get currently opened item
        self._load_on_open(item)

    def _load_on_open(self, item):
        vals = self.collection_tree.item(item, 'values')
        if len(vals) > 1 and vals[1] == 'Collection':
            te: ElementTreePage.TreeElementCollection = self._collection_registry[vals[0]]
            if not te.loaded:
                te.loaded = True
                thread = threading.Thread(target=self._load_next_page, args=(te,), daemon=True)
                thread.start()

    def _on_element_click(self, event):
        """Triggered when a tree item is opened (expanded)."""

        item = self.collection_tree.focus()  # Get currently opened item
        self._load_on_open(item)
        vals = self.collection_tree.item(item, 'values')
        if len(vals) < 2:
            return
        if vals[1] == 'LOAD':
            te: ElementTreePage.TreeElementCollection = self._collection_registry[vals[0]]
            te.loaded = True
            thread = threading.Thread(target=self._load_next_page, args=(te,), daemon=True)
            thread.start()
            self.collection_tree.delete(item)
        elif vals[1] == 'NewElement':
            te: ElementTreePage.TreeElementCollection = self._collection_registry[vals[0]]
            self.create_new(te.col)
        elif vals[1] == 'Element':
            if  hasattr(self, 'select_element'):
                parent = self.collection_tree.parent(item)
                col_id = self.collection_tree.item(parent, 'values')[0]
                te: ElementTreePage.TreeElementCollection = self._collection_registry[col_id]
                if self._element_type == 'Reaction':
                    element = self.instance.get_reaction(int(vals[0]))
                else:
                    element = self.instance.get_sample(int(vals[0]))
                self.select_element(te.col, element)
    
    def create_new(self, collection):
        pass


import itertools
import os
import tkinter as tk
from PIL import Image, ImageTk

__all__ = ['IconManager']

class LoadingGIF:
    def __init__(self, root):
        self.gif_path = os.path.join(os.path.dirname(__file__), 'fixed/loading.gif')
        self.root = root

        # Load the GIF
        self.image = Image.open(self.gif_path)
        self.frames = []
        self._call_id = None

        # Extract all frames
        try:
            while True:
                frame = self.image.copy().resize((16, 16), Image.LANCZOS)
                self.frames.append(ImageTk.PhotoImage(frame))
                self.image.seek(len(self.frames))  # Move to next frame
        except EOFError:
            pass  # End of frames

        self.current_frame = itertools.cycle(self.frames)

        # Display first frame
        self.labels = []

    def add_label(self, label: tk.Widget):
        self.labels.append(label)

    def _animate(self):
        """Update the GIF frame repeatedly"""
        self.labels = [l for l in self.labels if l.winfo_ismapped()]
        for l in self.labels:
            l.config(image=next(self.current_frame))
        self._call_id =  self.root.after(20, self._animate)  #  Adjust delay based on GIF speed

    def start(self):
        self._animate()

    def stop(self):
        self.root.after_cancel(self._call_id)
        self.labels = []



class IconManager():
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.load_icons()
        return cls._instance

    def load_icons(self):
        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/folder.svg.png')
        self.FOLDER_ICON = ImageTk.PhotoImage(Image.open(_icon_path).resize((16, 16)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/flask.svg.png')
        self.REACTION_ICON = ImageTk.PhotoImage(Image.open(_icon_path).resize((14, 14)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/plus.svg.png')
        self.PLUS_ICON = ImageTk.PhotoImage(Image.open(_icon_path).resize((14, 14)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/arrow-left-circle-fill.svg.png')
        self.BACK_ICON = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/arrow-right-circle-fill.svg.png')
        self.FORWARD_ICON = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/sample.png')
        self.SAMPLE_ICON = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/trash.png')
        self.TRASH = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/chemotion-full.png')
        self.CHEMOTION = ImageTk.PhotoImage(Image.open(_icon_path).resize((104 // 3,61 // 3)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/no_fav.png')
        self.NO_FAVORITE = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/fav.png')
        self.FAVORITE = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/R.png')
        self.R = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/python.png')
        self.PYTHON = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))

        _icon_path = os.path.join(os.path.dirname(__file__), 'fixed/info.png')
        self.INFO = ImageTk.PhotoImage(Image.open(_icon_path).resize((20,20)))


import csv  # Import the CSV module to read and write CSV files
import random
import sys  # Import sys to access command-line arguments

# Define the number of variations to generate
NUMBER_OF_VARIATIONS = 11

print(f'Generating {NUMBER_OF_VARIATIONS} random  variables')

# List to store the CSV data
values = []

# Read the input CSV file passed as the first command-line argument
print(f'Reading {sys.argv[1]}')
with open(sys.argv[1]) as csvfile:
    reader = csv.reader(csvfile)  # Create a CSV reader object
    for row in reader:
        values.append(list(row))  # Convert each row to a list and store it

# Calculate context-related values
print(f'Calculating context')
row_count = len(values)  # Total number of rows in the CSV
number_of_options = len(values[0]) - 2  # Number of available options per column (excluding first two columns)

print('Number of variables: ' + str(row_count))
print('Number of values per row: ' + str(number_of_options))

# Initialize results dictionary with VARIABLE and UNIT lists
# Initialize results list with headers and units
# [
# [VARIABLE, UNIT, Variation.1, Variation.2, ...]   First Row
# [S:xx, mol, 0.1, 0.2, ...]                        Second Row
# ...                                               ... nth row
# ]
results = [['VARIABLE', 'UNIT']] + [[row[0], row[1]] for i, row in enumerate(values)]


# Generate variations and store them in the results list
print(f'Preparing Variations')
for i in range(NUMBER_OF_VARIATIONS):
    results[0].append(f'Variation.{i}')  # Append variation headers
    for r_i in range(row_count):
        # Select a variation value based on a random choice
        results[r_i + 1].append(random.choice(values[r_i][2:]))

# Write the results to the output CSV file
print(f'Writing {sys.argv[2]}')
with open(sys.argv[2], 'w+') as csvfile:
    csvfile.write('\n'.join((','.join(row) for row in results)))  # Format and write CSV content



import tkinter as tk
from ChemDoE.icons import IconManager
from ChemDoE.login_manager import LoginManager
from ChemDoE.utils.page_manager import PageManager





def run():
    root = tk.Tk()

    photo = IconManager().CHEMOTION
    root.iconphoto(True, photo)

    root.title("Chemotion DoE")
    w, h = root.winfo_screenwidth(), root.winfo_screenheight()
    root.geometry("%dx%d+0+0" % (max(w, 700), max(h, 600)))
    pm = PageManager(root)
    pm.start_page(LoginManager())


if __name__ == '__main__':
    run()



def validate_results(data: dict) -> list[str]:
    errors = []
    if 'VARIABLE' not in data:
        errors.append(f'VARIABLE must be a key in the result JSON!')
    if 'UNIT' not in data:
        errors.append(f'UNIT must be a key in the result JSON!')
    array_length = None
    for k, val in data.items():
        if not isinstance(val, list):
            errors.append(f'{k} must be an array!')
            continue

        if array_length is None:
            array_length = len(val)
        elif len(val) != array_length:
            errors.append(f'{k} must have {array_length} elements!')

        if k not in ['UNIT', 'VARIABLE']:
            try:
                for i, x in enumerate(val):
                    val[i] = float(x)
            except ValueError:
                errors.append(f'{k} must be an array of numbers!')

    return errors


import json
import os
import re
import threading
from pathlib import Path
from subprocess import Popen, PIPE
import tkinter as tk
from tkinter import ttk, messagebox

from chemotion_api import Reaction

from ChemDoE.config import ConfigManager
from ChemDoE.icons import IconManager
from ChemDoE.registery.loader import read_csv, read_json
from ChemDoE.registery.result_validater import validate_results
from ChemDoE.utils.pages import ScrollableFrame


class ExecuteManager(tk.Toplevel):

    def __init__(self, root, reaction: Reaction):
        super().__init__(root)
        self._root = root
        self.title("Run")
        self.geometry("600x300")
        self.sf = ScrollableFrame(self, horizontal=True)
        self.sf.pack(fill="both", expand=True, padx=10, pady=10)
        self._label = ttk.Label(self.sf.scrollable_frame, text="Running...", anchor="w")
        self._label.pack(fill="x", padx=10, pady=10)
        self._reaction = reaction
        self._errors = []

    def run(self, script, values):
        fp = Path(script['file'])
        ft = script['input'].lower()
        out_ft = script['output'].lower()
        file_name = re.sub(r'[^A-Za-z]', '_', script['name']) + f'.{ft}'

        file_path = str(fp.parent / file_name)

        if ft == 'json':
            with open(file_path, 'w+') as f:
                f.write(json.dumps(values, indent=4))
        else:
            res = []
            for key, val in values.items():
                res.append(','.join([key] + [str(x) for x in val]))
            with open(file_path, 'w+') as f:
                f.write('\n'.join(res))
        out_file_path = str(fp.parent / f'out_{file_name}.{out_ft}')
        cmd = [script['interpreter'], script['file'], file_path, out_file_path]

        p = Popen(cmd, stdout=PIPE,
                  stderr=PIPE,
                  text=True,  # Ensures output is in string format instead of bytes
                  bufsize=1,  # Line-buffered output
                  universal_newlines=True,  # Handles newlines properly across platforms
                  cwd=fp.parent
                  )
        lines = ""
        for line in iter(p.stdout.readline, ''):
            lines += line
            self._root.after(0, lambda: self._label.config(text=lines))
        os.remove(file_path)
        self._root.after(10, self.load_results, out_file_path, out_ft)

    def load_results(self, out_file_path, out_ft):
        tree = ttk.Treeview(self._label.master)
        self._label.config(text=self._label.cget('text') + "\nResults")
        tree.pack(expand=True, fill="x", padx=10, pady=10)
        if out_ft == 'json':
            self._data = self.load_json(tree, out_file_path)
        elif out_ft == 'csv':
            self._data = self.load_csv(tree, out_file_path)
        if self._data is not None:
            self._root.after(10, self.sf.update_sc_view)
            ttk.Button(tree.master, text="Transfer variations", image=IconManager().CHEMOTION, compound="left",
                       command=self.transfer_variations).pack(anchor="w")
            self.sf.update_sc_view()
        os.remove(out_file_path)

    def transfer_variations(self):
        self._label.config(text="Transferring...")

        def save():
            self._reaction = ConfigManager().chemotion.get_reaction(self._reaction.id)
            for v in self._reaction.variations[:]:
                if v.notes.startswith('ChemDoE: '):
                    self._reaction.variations.remove(v)
            self._reaction.save(True)
            for v_name, val in self._data.items():
                if v_name not in ['UNIT', 'VARIABLE']:
                    variation = self._reaction.variations.add_new()
                    variation.notes = f'ChemDoE: <<{v_name}>>'
                    for materials in ['starting_materials', 'reactants', 'products', 'solvents']:
                        letter = materials.upper()[0]
                        for mat in getattr(variation, materials):
                            key = f"{letter}:{mat.sample.id}"
                            idx = self._data['VARIABLE'].index(key)
                            mat.set_quantity(float(val[idx]), str(self._data['UNIT'][idx]))
                    for key, prop in variation.properties.items():
                        key = key.capitalize()
                        idx = self._data['VARIABLE'].index(key)
                        prop['value'] = float(val[idx])
                        prop['unit'] = str(self._data['UNIT'][idx])
            self._reaction.save()
            self._root.after(0, lambda: self._label.config(text="Transferred"))

        threading.Thread(target=save).start()

    def _validate_json_results(self, data):
        self._errors = validate_results(data)

        return len(self._errors) == 0

    def load_json(self, tree, file_path):
            return self._render_tree(tree, read_json(file_path))

    def load_csv(self, tree, file_path):
            return self._render_tree(tree, read_csv(file_path))

    def _render_tree(self, tree, reader):
        if not self._validate_json_results(reader):
            messagebox.showerror('Invalid JSON!', '\n'.join(f' - {x}' for x in self._errors))
            return None
        headers = list(reader.keys())

        headers.remove('VARIABLE')
        headers.remove('UNIT')
        headers = ['VARIABLE', 'UNIT'] + headers
        tree["columns"] = headers
        tree["show"] = "headings"  # Hide default first column
        new_height = 1
        # Set up column headers
        for col in headers:
            tree.heading(col, text=col)
            tree.column(col, anchor="w", width=100)

        # Insert data into the treeview
        for i in range(len(reader[headers[0]])):
            new_height += 1
            tree.insert("", "end", values=[reader[head][i] for head in headers])

        tree.configure(height=new_height)
        return reader



from typing import Literal

import click
from ChemDoE.main import run as run_ui
from ChemDoE.registery import register_r_doe_script, register_python_doe_script

FileFormat = ['csv', 'json']


@click.group()
def cli():
    pass


@cli.command(help="Runs the CheDoE UI")
def run():
    run_ui()


@cli.command(
    help="Executes a R script with test DoE Table and adds it to the run scripts of ChemDoE if the results are in the correct format")
@click.option('--r-script-path', '-s', prompt='Path to the R script', help='Path to the R script')
@click.option('--output-format', '-o', type=click.Choice(FileFormat), prompt='Output file format',
              default=FileFormat[0], help='Output file format must be csv or json')
@click.option('--input-format', '-i', type=click.Choice(FileFormat), prompt='Input file format', default=FileFormat[0],
              help='Input file format must be csv or json')
def add_r_script(r_script_path: str, input_format: Literal['csv', 'json'] , output_format: Literal['csv', 'json']):
    register_r_doe_script(r_script_path, input_format, output_format)


@cli.command(
    help="Executes a Python script with test DoE Table and adds it to the run scripts of ChemDoE if the results are in the correct format")
@click.option('--python-script-path', '-s', prompt='Path to the Python script', help='Path to the Python script')
@click.option('--output-format', '-o', type=click.Choice(FileFormat), prompt='Output file format',
              default=FileFormat[0], help='Output file format must be csv or json')
@click.option('--input-format', '-i', type=click.Choice(FileFormat), prompt='Input file format', default=FileFormat[0],
              help='Input file format must be csv or json')
def add_python_script(python_script_path: str, input_format: Literal['csv', 'json'] , output_format: Literal['csv', 'json']):
    register_python_doe_script(python_script_path, input_format, output_format)


if __name__ == '__main__':
    cli()



from tkinter import ttk
import tkinter as tk
import tkinter.font as tkFont

from chemotion_api import Reaction
from chemotion_api.labImotion.items.options import FieldType, UNITS

from ChemDoE.config import ConfigManager
from ChemDoE.script_manager import ScriptOrganizer
from ChemDoE.utils.pages import ToolBarPage


class DoEPage(ToolBarPage):
    def __init__(self, reaction: Reaction):
        super().__init__()
        self._run_manager = None
        self.reaction = reaction
        self._name = f'{self.reaction.short_label} {self.reaction.name}'
        self._rows = None
        self._additional_fields = []
        self._addable_fields = None
        self._addable_field_dropdown = None
        self._table = None
        self._columns = 3
        self._columnes_var = tk.IntVar(value=self._columns)
        self._columnes_var.trace_add("write", self._columns_change)
        self._entries = {}
        self._values = {}
        self._save_after_id = None

    def render(self, container: ttk.Frame):
        super().render(container)

        ttk.Label(container, text=self._name, style='Header.TLabel',
                  font=ConfigManager.header_font).pack(fill='x')
        button_frame = tk.Frame(container)
        button_frame.pack(fill='x')
        
        self._run_manager = ScriptOrganizer(self._page_manager, button_frame, self.reaction)
        self._run_manager.pack(side='left', padx=5, pady=5)
        ttk.Label(button_frame, text="# columns:").pack(side='left', padx=5, pady=5)
        ttk.Entry(button_frame, textvariable=self._columnes_var, validate="all", width=4,
                  validatecommand=self._page_manager.validators["numeric"]).pack(side='left', padx=5, pady=5)
        ttk.Button(button_frame, text="Reshape", command=self._update_table).pack(side='left', padx=5, pady=5)
        ttk.Button(button_frame, text="Save Template", command=lambda *x: self._save_template(self._name)).pack(
            side='left',
            padx=5,
            pady=5)
        selected_template = ttk.Combobox(button_frame, values=[], state="readonly")
        selected_template.bind("<Button-1>", lambda *x: self._on_dropdown_template(selected_template))
        selected_template.pack(side='left', padx=5, pady=5)
        ttk.Button(button_frame, text="Load Template", command=lambda *x: self._load_templte(selected_template)).pack(
            side='left', padx=5, pady=5)

        self._table = ttk.Frame(container, style='Table.TFrame')
        self._table.pack(fill='x')
        

        values = ConfigManager().load_doe(self._name)
        if values and 'values' in values and 'template' in values:
            self._values = values['values']
            self._execute_template(values['template'])
        self._run_manager.values = self._values
        self._update_table()

    def set_style(self, style: ttk.Style):
        ScriptOrganizer.set_style(style)
        style.configure('Table.TFrame', background='white')
        style.configure('Header.TLabel', background='white')
        style.configure('TableHeader.TLabel', borderwidth=1, relief='solid', background='white', padding=10)
        style.configure('TableRowHeader.TLabel', borderwidth=1, relief='solid', background='white', padding=5)
        style.configure('TableInnerHeader.TLabel', background='#0091ea', padding=(10, 2))
        style.configure('AddButton.TButton',  padding=(10, 2))
        style.configure('TableInnerRemove.TButton',  padding=(2, 2))

        style.map('AddButton.TButton',
                  foreground=[('!active', 'white'), ('pressed', 'white'), ('active', '#333333')],
                  background=[('!active', '#5cb85c'), ('pressed', '#5cb85c'), ('active', '#5cb85c')]
                  )

        style.map('TableInnerRemove.TButton',
                  foreground=[('!active', 'white'), ('pressed', 'white'), ('active', '#333333')],
                  background=[('!active', '#d9534f'), ('pressed', '#d9534f'), ('active', '#d9534f')]
                  )

    def _get_validation_templates(self):
        v = self.reaction.variations.add_new()
        self.reaction.variations.remove(v)
        return v

    def _get_addable_fields(self):
        self._addable_fields = [s for s in ConfigManager().all_additional_fields() if s not in self._additional_fields]
        return self._addable_fields

    def _update_table(self, *_args):
        self._clean_table()
        ttk.Frame(self._table, style='TableHeader.TLabel').grid(row=0, column=0)
        ttk.Label(self._table, text='Units', style='TableHeader.TLabel', anchor='center').grid(row=0, column=1,
                                                                                               sticky="ew")
        for i in range(2, self._columns + 2):
            if i == 2:
                text = 'Min'
            elif i - 1 == self._columns:
                text = 'Max'
            else:
                text = '     '
            ttk.Label(self._table, text=text, style='TableHeader.TLabel', anchor='center').grid(row=0, column=i,
                                                                                                sticky="ew")
        row = 0
        validation = self._get_validation_templates()

        for key, title in (
                ("starting_materials", "Starting Material"), ("reactants", "Reactants"), ("products", "Products"),
                ("solvents", "Solvents")):
            row = self._render_mat_rows(row + 1, validation, key, title)
        row += 1
        ttk.Label(self._table, text='Properties', style='TableInnerHeader.TLabel', justify='left').grid(row=row,
                                                                                                        column=0,
                                                                                                        sticky="ew",
                                                                                                        columnspan=self._columns + 2)
        for k, p in validation.properties.items():
            row += 1
            text = k.title()
            units = [p['unit']]
            self._render_row(text, row, text, units)

        row += 1
        ttk.Label(self._table, text='Segment properties', style='TableInnerHeader.TLabel', justify='left').grid(row=row,
                                                                                                                column=0,
                                                                                                                sticky="ew",
                                                                                                                columnspan=self._columns + 2)
        row += 1
        add_frame = ttk.Frame(self._table, style='TableHeader.TLabel')
        add_frame.grid(row=row, column=0, sticky="ew", columnspan=self._columns + 2)

        fields = [f"{s[0].label}: {s[1].label}: {s[2].label}" for s in self._get_addable_fields()]
        self._addable_field_dropdown = ttk.Combobox(add_frame, values=fields, state="readonly")

        font = tkFont.Font(font=self._addable_field_dropdown.cget("font"))
        if len(fields) > 0:
            max_width = max(font.measure(option) for option in fields)  # Get max text width in pixels
            self._addable_field_dropdown.config(width=max_width // font.measure("0") + 2)

        self._addable_field_dropdown.pack(side='left', padx=5, pady=5)
        ttk.Button(add_frame, text='Add properties', style='AddButton.TButton', command=self._add_addable_field).pack(
            side='left', padx=5, pady=5)

        for s in self._additional_fields:
            row += 1
            text = f"{s[0].label}: {s[1].label}: {s[2].label}"
            row_id = '__'.join(self._field_to_key(s))
            units = '-'
            if s[2].field_type == FieldType.SYSTEM_DEFINED:
                units = [x.replace('_2', '²').replace('_3', '³').replace('2', '²').replace('3', '³').replace('_', '/')
                         for x in UNITS[s[2].option_layers]]
            button = ttk.Button(self._table, text="🗑", style='TableInnerRemove.TButton', width=2,
                                command=lambda x=s: self._remove_addable_field(x))
            self._render_row(row_id, row, text, units, button)

        self._clean_values()
        self._save()

    @staticmethod
    def _field_to_key(field_tuple):
        return [field_tuple[0].identifier, field_tuple[1].key, field_tuple[2].field]

    def _render_mat_rows(self, row, validation, key, title):
        mats = getattr(validation, key)
        if len(mats) == 0:
            return row - 1
        ttk.Label(self._table, text=title, style='TableInnerHeader.TLabel', justify='left').grid(row=row, column=0,
                                                                                                 sticky="ew",
                                                                                                 columnspan=self._columns + 2)
        for i in mats:
            row += 1
            text = f"{title[0]}:{i.sample.id}: {i.sample.short_label}: {i.sample.molecule['cano_smiles']}"
            units = [x.value for x in i.potential_units()]
            self._render_row(f'{title[0]}:{i.sample.id}', row, text, units)
        return row

    def _remove_addable_field(self, row):
        self._additional_fields.remove(row)
        self._update_table()

    def _add_addable_field(self, *_args):
        idx = self._addable_field_dropdown.current()
        if idx == -1:
            return
        self._additional_fields.append(self._addable_fields[idx])
        self._update_table()

    def _on_table_entry_change(self, row_id, col):
        self._values[row_id][col] = self._entries[row_id][col].get()
        if self._save_after_id is not None:
            self._page_manager.root.after_cancel(self._save_after_id)
        self._save_after_id = self._page_manager.root.after(500, self._save)

    def _on_table_unit_change(self, row_id):
        self._on_table_entry_change(row_id, col=0)

    def _render_row(self, row_id, row, text, units, button=None):
        label_frame = ttk.Frame(self._table)
        label_frame.grid(row=row, column=0, sticky="ew")
        ttk.Label(label_frame, text=text, style='TableRowHeader.TLabel').pack(fill='x')
        unit_var = tk.StringVar(value=self._get_value_unit(row_id, units))
        unit_sel = ttk.Combobox(self._table, values=units, state="readonly", textvariable=unit_var)
        unit_var.trace_add("write", lambda _a, _b, _c: self._on_table_unit_change(row_id))
        unit_sel.grid(row=row, column=1, sticky="ew")
        self._entries[row_id] = [unit_var]
        for col_idx in range(self._columns):
            col = col_idx + 2
            ev = tk.DoubleVar(value=self._get_value(row_id, col_idx))
            self._entries[row_id].append(ev)
            entry = ttk.Entry(self._table, textvariable=ev, validate="all",
                              validatecommand=self._page_manager.validators["float"])
            entry.grid(row=row,
                       column=col,
                       sticky="ns")
            ev.trace_add("write", lambda _a, _b, _c, i=col_idx: self._on_table_entry_change(row_id, i + 1))

        if button:
            button.grid(row=row, column=self._columns + 2)

    def _clean_table(self):
        for widget in self._table.winfo_children():
            widget.destroy()
            del widget

    def _get_value(self, row, col):
        col += 1
        if row not in self._values:
            self._values[row] = [''] + [0.0] * self._columns
        while len(self._values[row]) <= col:
            self._values[row].append(0.0)
        return self._values[row][col]

    def _get_value_unit(self, row, default_units):
        value = self._get_value(row, -1)
        if value not in default_units:
            self._values[row][0] = default_units[0]
        return self._values[row][0]

    def _columns_change(self, *_args):
        try:
            a = self._columnes_var.get()
            if a > 0:
                self._columns = self._columnes_var.get()
        except tk.TclError:
            return

    def _prepare_template(self):
        return {
            'size': self._columns,
            'fields': [self._field_to_key(s) for s in self._additional_fields]
        }

    def _save_template(self, name):
        template = self._prepare_template()
        ConfigManager().save_template(template, name)

    def _load_templte(self, select_template):
        selected_template = select_template.get()
        if selected_template == '':
            return
        template = ConfigManager().load_template(selected_template)
        self._execute_template(template)

    def _execute_template(self, template):
        self._columns = template['size']
        self._columnes_var.set(template['size'])
        self._additional_fields.clear()
        for field in template['fields']:
            for potential_field in self._get_addable_fields():
                if field[0] == potential_field[0].identifier and field[1] == potential_field[1].key and field[2] == \
                        potential_field[2].field:
                    self._additional_fields.append(potential_field)

        self._update_table()

    @staticmethod
    def _on_dropdown_template(selected_template):
        selected_template.config(values=[x.name[:-5] for x in ConfigManager().load_templates()])

    def _clean_values(self):
        for key in list(self._values.keys()):
            if key not in self._entries.keys():
                del self._values[key]
            else:
                self._values[key] = self._values[key][:self._columns + 1]

    def _save(self):
        ConfigManager().save_doe({'values':self._values, 'template': self._prepare_template()}, self._name)



import configparser
import glob
import json
import os.path
import shutil
import sys
import uuid
from pathlib import Path
from typing import Optional

from chemotion_api import Reaction
from chemotion_api.labImotion.items.options import FieldType
from platformdirs import user_config_dir
from chemotion_api import Instance
from requests import RequestException

config_dir = Path(user_config_dir("ChemDoE"))
templates_dir = config_dir / "templates"
script_dir = config_dir / "scripts"
doe_dir = config_dir / "doe"
config_path = config_dir / "config.ini"


class ConfigManager:
    _instance = None
    header_font = ('Arial', 16, 'bold')
    small_font = ('Arial', 10)
    normal_font = ('Arial', 12)

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.config = configparser.ConfigParser()
            cls._instance.read()
            cls._instance._favorites_reactions = None
            cls._instance._instance_key = None
            cls._instance._chemotion = None
            cls._instance._segments = None
            cls._instance.python_interpreters = []
            cls._instance.r_interpreters = []
            cls._instance.find_all_python_interpreters()
            cls._instance.find_r_interpreters()
        return cls._instance

    @property
    def chemotion(self) -> Optional[Instance]:
        return self._chemotion

    @chemotion.setter
    def chemotion(self, value: Optional[Instance]):
        self._chemotion = value

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = ConfigManager()
        return cls._instance

    @classmethod
    def read(cls):
        if os.path.exists(config_path):
            cls.instance().config.read(config_path)
        else:
            cls.instance()._set_defaults()

    @classmethod
    def set(cls, section, option, value, commit: bool = True):
        if section not in cls.instance().config.sections():
            cls.instance().config.add_section(section)
        cls.instance().config.set(section, option, value)
        if commit:
            cls.instance().save()
            cls.instance().config = configparser.ConfigParser()
            cls.instance().read()

    @classmethod
    def get(cls, section, option, default=None):
        try:
            return cls.instance().config.get(section, option)
        except (configparser.NoOptionError, configparser.NoSectionError):
            return default

    def logout(self):
        self.set("Chemotion", "host", '', commit=False)
        self.set("Chemotion", "user", '', commit=False)
        self.set("Chemotion", "token", '')
        self.chemotion = None
        self._favorites_reactions = None

    @property
    def favorites_with_names(self):
        if self.chemotion is None:
            self._favorites_reactions = []
        elif self._favorites_reactions is None:
            self._favorites_reactions = []
            for fav_id in self._load_favorites():
                if not self._load_add_reaction_fav(fav_id):
                    self.remove_from_favorites(fav_id)
        return self._favorites_reactions

    def _load_add_reaction_fav(self, reaction_id: int, reaction: Optional[Reaction] = None) -> bool:
        if reaction is None:
            try:
                reaction = self.chemotion.get_reaction(reaction_id)
            except RequestException:
                return False
        self._favorites_reactions.append((reaction_id, f"{reaction.short_label}: {reaction.name}"))
        return True

    @property
    def favorites(self):
        return [x[0] for x in self.favorites_with_names]

    def _load_favorites(self):
        self._instance_key = f"{self.get('Last', 'Host')}_{self.get('Last', 'User')}"
        return [int(id) for id in json.loads(self.get(self._instance_key, "Reaction", '[]'))]

    def add_to_favorites(self, reaction: Reaction):
        if reaction.id in self.favorites:
            return
        self._load_add_reaction_fav(reaction.id, reaction)
        self.favorites_with_names.sort(key=lambda x: int(x[0]))
        self.set(self._instance_key, "Reaction", json.dumps(self.favorites))

    def remove_from_favorites(self, reaction: Reaction | int):
        try:
            if isinstance(reaction, Reaction):
                reaction = reaction.id
            idx = self.favorites.index(int(reaction))
            del self._favorites_reactions[idx]
            self.set(self._instance_key, "Reaction", json.dumps(self.favorites))
        except ValueError:
            pass

    def _set_defaults(self):
        self.config["DEFAULT"] = {
            "AppName": "ChemDoE",
            "Debug": "False"
        }

        self.config["Chemotion"] = {
            "Host": "",
            "User": "",
            "Token": "",
        }

        self.config["Last"] = {
            "Host": "",
            "User": "",
            "Remember": "0",
        }

        self.config["Favorites"] = {
            "Reaction": "[]"
        }
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        self.save()

    def save(self):
        with open(config_path, "w") as configfile:
            self.config.write(configfile)

    def all_additional_fields(self):
        if self._segments is None:
            segments = self.chemotion.generic_manager().load_all_segments()
            self._segments = []
            for segment in segments:
                if segment.identifier is None:
                    segment.identifier = uuid.uuid4().__str__()
                    segment.save()
                if segment._element_klass.name == 'reaction' and segment.is_active:
                    for l in segment.properties.layers:
                        for f in l.fields:
                            if f.field_type in [FieldType.SYSTEM_DEFINED, FieldType.FORMULA_FIELD, FieldType.INTEGER]:
                                self._segments.append((segment, l, f))
        return self._segments

    def save_doe(self, template, name):
        os.makedirs(doe_dir, exist_ok=True)
        with open(doe_dir / (name + '.json'), "w+") as configfile:
            configfile.write(json.dumps(template))

    def load_doe(self, name, ):
        file_path = doe_dir / (name + '.json')
        if not os.path.exists(file_path):
            return None
        with open(file_path, "r") as configfile:
            return json.loads(configfile.read())

    def load_scripts(self):
        file_path = script_dir / 'config.json'
        if not os.path.exists(file_path):
            self.save_scripts([])
        with open(file_path, "r") as configfile:
            return json.loads(configfile.read())['runs'] + self._generate_default_scripts()

    def save_scripts(self, content: list[dict]):
        file_path = script_dir / 'config.json'
        os.makedirs(script_dir, exist_ok=True)
        with open(file_path, "w+") as configfile:
            configfile.write(json.dumps({'runs': [x for x in content if not x['id'].startswith('__default_')]}))

    def save_template(self, template, name):
        os.makedirs(templates_dir, exist_ok=True)
        with open(templates_dir / (name + '.json'), "w+") as configfile:
            configfile.write(json.dumps(template))

    def load_template(self, name):
        with open(templates_dir / (name + '.json'), "r") as configfile:
            return json.loads(configfile.read())

    def load_templates(self):
        os.makedirs(templates_dir, exist_ok=True)
        return [file for file in templates_dir.iterdir() if file.is_file()]

    def find_all_python_interpreters(self):
        interpreters = set()
        interpreters.update(self.find_python_interpreters())
        interpreters.update(self.find_python_in_common_locations())

        self.python_interpreters = list(interpreters)
        self.python_interpreters.sort(key=lambda x: len(x))
        if len(self.python_interpreters) == 0:
            from tkinter import messagebox
            messagebox.showwarning("Python not installed", "Please install a Python interpreter!")

    @staticmethod
    def find_python_interpreters():
        possible_interpreters = ["python", "python3", "python3.10", "python3.11", "python3.12", "python3.13"]
        found_interpreters = []

        for interpreter in possible_interpreters:
            path = shutil.which(interpreter)
            if path and path not in found_interpreters:
                found_interpreters.append(path)

        return found_interpreters

    @staticmethod
    def find_python_in_common_locations():
        paths = [
            "C:\\Python*\\python.exe",  # Common Windows installations
            "C:\\Users\\*\\AppData\\Local\\Programs\\Python\\Python*\\python.exe",
            "/usr/bin/python*",  # Linux/macOS system-wide
            "/usr/local/bin/python*",
            "/opt/python*",
        ]

        found = []
        for pattern in paths:
            found.extend([x for x in glob.glob(pattern) if not x.endswith("-config")])

        return found

    @staticmethod
    def find_win_r_executables() -> list[str]:
        possible_paths = [
            os.path.join(os.getenv("ProgramFiles"), "R"),  # Standard install location
            os.path.join(os.getenv("ProgramFiles(x86)"), "R"),  # 32-bit install location
            os.path.join(os.getenv("LOCALAPPDATA"), "Programs", "R"),  # Some custom installs
            "C:\\R"  # Occasionally used custom path
        ]

        found_executables = []

        # Search for R.exe in common paths
        for base_path in possible_paths:
            if base_path and os.path.exists(base_path):
                found_executables.extend(glob.glob(os.path.join(base_path, "**", "Rscript.exe"), recursive=True))

        # Search in system PATH
        for path in os.getenv("PATH", "").split(os.pathsep):
            r_exe = os.path.join(path, "Rscript.exe")
            if os.path.exists(r_exe):
                found_executables.append(str(r_exe))

        return found_executables

    def find_r_interpreters(self):
        r_path = shutil.which("Rscript")
        if not r_path:
            r_path = []
        else:
            r_path = [r_path]

        if os.name == "nt":
            r_path += self.find_python_interpreters()

        if len(r_path) == 0:
            from tkinter import messagebox
            messagebox.showwarning("R not installed", "Please install a R interpreter!")

        self.r_interpreters = r_path

    def _generate_default_scripts(self):
        results = []
        src_dir = Path(__file__).parent / 'examples'
        os.makedirs(script_dir, exist_ok=True)
        if len(self.python_interpreters) >= 0:
            for i, filename in enumerate(os.listdir(src_dir)):
                src = src_dir / filename
                if src.suffix != '.py':
                    continue
                dst = script_dir / filename
                if not dst.exists():
                    shutil.copyfile(src, dst)
                results.append({
                    'id': f'__default_P{i}',
                    'name': f'[Default] {filename}',
                    'file_type': 'Python',
                    'input': filename.split('_')[0].upper(),
                    'output': filename.split('_')[0].upper(),
                    'file': dst.__str__(),
                    'interpreter': self.python_interpreters[0]
                })
        if len(self.r_interpreters) >= 0:
            for i, filename in enumerate(os.listdir(src_dir)):
                src = src_dir / filename
                dst = script_dir / filename

                if src.suffix != '.R':
                    continue
                if not dst.exists():
                    shutil.copyfile(src, dst)
                results.append({
                    'id': f'__default_R{i}',
                    'name': f'[Default] {filename}',
                    'file_type': 'R',
                    'input': filename.split('_')[0].upper(),
                    'output': filename.split('_')[0].upper(),
                    'file': str(dst),
                    'interpreter': self.r_interpreters[0]
                })


        return results



from tkinter import ttk
from abc import ABC, abstractmethod
from typing import Optional
import tkinter as tk


def validate_float(P):
    if validate_numeric(P):
        return True
    try:
        float(P)
        return True
    except ValueError:
        return False


def validate_numeric(P):
    """Allow only digits"""
    return P.isdigit() or P == ""

class PageManager:
    def __init__(self, root: tk.Tk):
        self.validators = {}
        self._root = root
        self.page: Optional[Page] = None
        self.outer_frame = ttk.Frame(root, style="Outer.TFrame")
        self.outer_frame.pack(fill="both", expand=True)
        self._history: list[Page] = []
        self._history_idx = 0
        self._is_running = False

    @property
    def root(self) -> tk.Tk:
        return self._root

    @property
    def history(self):
        return self._history


    def can_go_back(self, n: int = 1) -> bool:
        return self._history_idx > n - 1

    def can_go_forward(self, n: int = 1) -> bool:
        return self._history_idx < len(self._history) - n

    def start_page(self, page: "Page"):
        self._history = []
        if not self._is_running:
            self.validators["numeric"] = (self._root.register(validate_numeric), "%P")
            self.validators["float"] = (self._root.register(validate_float), "%P")
        self.set_page(page)
        if not self._is_running:
            self._root.mainloop()
        self._is_running = True



    def set_page(self, page: "Page"):
        self._history = self._history[:self._history_idx + 1]
        self._history.append(page)
        self.go_to_history_idx(len(self._history) - 1)

    def go_back(self):
        if not self.can_go_back():
            raise RuntimeError("Cannot go back to previous page!")
        return self.go_to_history_idx(self._history_idx - 1)

    def go_forward(self):
        if not self.can_go_forward():
            raise RuntimeError("Cannot go forward to next page!")
        return self.go_to_history_idx(self._history_idx + 1)

    def go_to_history_idx(self, idx: int = -1):
        page = self._history[idx]
        self._history_idx = idx
        style = ttk.Style()
        style.theme_use("clam")

        style.configure(".", font=("Arial", 12), background="#ffffff", foreground="black")
        style.map("TEntry", fieldbackground=[("focus", "white"), ("!focus", "lightgray")])

        if self.page:
            self.page.is_visible = False
            self.page.leave()
        self.page = page
        self.page.is_visible = True
        page.page_manager = self

        self._clear_frame()

        page.set_style(style)
        page.render(self.outer_frame)

    def _clear_frame(self):
        for widget in self.outer_frame.winfo_children():
            widget.destroy()
            del widget


class Page(ABC):

    def __init__(self):
        self.is_visible = False
        self._page_manager: Optional[PageManager] = None

    def leave(self):
        pass

    @property
    def page_manager(self) -> PageManager:
        return self._page_manager

    @page_manager.setter
    def page_manager(self, value: PageManager):
        if not isinstance(value, PageManager):
            raise TypeError(f"page_manager {type(value)} not supported")
        self._page_manager = value

    @abstractmethod
    def render(self, root: ttk.Frame):
        ...

    @abstractmethod
    def set_style(self, style: ttk.Style):
        ...



from tkinter import ttk
import tkinter as tk
from typing import Callable
from abc import ABC, abstractmethod, abstractproperty

from ChemDoE.utils.utils import clone_widget


class DragManager(ABC):

    def __init__(self):
        self._on_drag_start = lambda x: True
        self._on_drop_handler = []

    @property
    def on_drag_start(self) -> Callable[[str], bool]:
        return self._on_drag_start

    @on_drag_start.setter
    def on_drag_start(self, value: Callable[[str], bool]):
        self._on_drag_start = value

    def on_drop_in_target(self, target: tk.Widget, value: Callable[[str | tk.Widget, tk.Widget], bool]):
        self._on_drop_handler.append((target, value))


class DragManagerWidget(DragManager):
    def __init__(self, root: tk.Tk):
        super().__init__()
        self.drag_window = None
        self.dragged_item = None
        self.root = root

        self.key_pressed = False

    def set_elements(self, elements: list[ttk.Widget]):
        self.elements = elements

        for e in self.elements:
            self.bind_all_events(e)

    def bind_all_events(self, e: ttk.Widget):
        e.bind("<ButtonPress-1>", lambda event: self._on_start(e, event), True)
        e.bind("<B1-Motion>", lambda event: self._on_drag(e, event), True)
        e.bind("<ButtonRelease-1>", lambda event: self._on_drop(e, event), True)

    def _on_start(self, e: ttk.Widget, event):
        self.key_pressed = True
        self.root.after(100, self._start_dragging, e, event)

    def _start_dragging(self, e: ttk.Widget, event):
        if not self.key_pressed or not e.winfo_exists():
            return

        if not self._on_drag_start(e):
            return
        self.dragged_item = e
        self.drag_window = tk.Toplevel(self.root)
        self.drag_window.overrideredirect(True)

        self._drag_window_width = e.winfo_width()
        self._drag_window_height = e.winfo_height()

        self.drag_window.geometry(f"{self._drag_window_width}x{self._drag_window_height}+{event.x_root}+{event.y_root}")

        clone_widget(e, self.drag_window).pack(fill="both", expand=True)

        # that represents what is being dragged.
        pass

    def _on_drag(self, e, event):
        # you could use this method to move a floating window that
        # represents what you're dragging
        if self.drag_window:
            self.drag_window.geometry(f"{self._drag_window_width}x{self._drag_window_height}+{event.x_root}+{event.y_root}")

    def _on_drop(self, e, event):
        self.key_pressed = False
        if self.drag_window is None:
            return
        # find the widget under the cursor
        x, y = event.widget.winfo_pointerxy()
        self.drag_window.destroy()
        self.drag_window = None
        first_target = event.widget.winfo_containing(x, y)
        for goal_target, handler in self._on_drop_handler:
            target = first_target
            found = False
            while target:
                found |= target == goal_target
                if found:
                    break
                target = target.master
            if found:
                try:
                    handler(self.dragged_item, target)
                except:
                    pass
                return


class DragManagerTree(DragManager):
    def __init__(self, tree: ttk.Treeview, root: tk.Tk):
        super().__init__()
        self.drag_window = None
        self.tree = tree
        self.dragged_item = None
        self.root = root

        self.tree.bind("<ButtonPress-1>", self._on_start, True)
        self.tree.bind("<B1-Motion>", self._on_drag, True)
        self.tree.bind("<ButtonRelease-1>", self._on_drop, True)
        self.tree.configure(cursor="hand1")

        self.key_pressed = False
        self._on_drag_start = lambda x: True
        self._on_drop_handler = []

    def _on_start(self, event):
        self.key_pressed = True
        self.root.after(100, self._start_dragging, event)

    def _start_dragging(self, event):
        if not self.key_pressed or not self.tree.winfo_exists():
            return

        selection = self.tree.selection()[0]
        if not selection or not self._on_drag_start(selection):
            return
        self.dragged_item = selection
        dragged_text = self.tree.item(selection, "text")
        self.drag_window = tk.Toplevel(self.root)
        self.drag_window.overrideredirect(True)
        self.drag_window.geometry(f"100x30+{event.x_root}+{event.y_root}")

        label = tk.Label(self.drag_window, text=dragged_text, background="lightgray", relief="solid", borderwidth=0)
        label.pack(fill="both", expand=True)

        # that represents what is being dragged.
        pass

    def _on_drag(self, event):
        # you could use this method to move a floating window that
        # represents what you're dragging
        if self.drag_window:
            self.drag_window.geometry(f"100x30+{event.x_root}+{event.y_root}")

    def _on_drop(self, event):
        self.key_pressed = False
        if self.drag_window is None:
            return
        # find the widget under the cursor
        x, y = event.widget.winfo_pointerxy()
        self.drag_window.destroy()
        self.drag_window = None
        first_target = event.widget.winfo_containing(x, y)
        for goal_target, handler in self._on_drop_handler:
            target = first_target
            found = False
            while target:
                found |= target == goal_target
                if found:
                    break
                target = target.master
            if found:
                try:
                    handler(self.dragged_item, target)
                except:
                    pass
                return



import threading
from tkinter import ttk, messagebox
from typing import Optional
import tkinter as tk

from chemotion_api import Instance, Sample
from chemotion_api.collection import Collection
from requests import RequestException

from ChemDoE.config import ConfigManager
from ChemDoE.icons import LoadingGIF
from ChemDoE.utils.pages import ToolBarPage, ScrollableFrame


class NewSample(ToolBarPage):
    def __init__(self, collection: Collection, sample: Optional[Sample] = None):
        super().__init__()
        self._save_btn = None
        self._collection = collection
        self._is_new = sample is None

        if self._is_new:
            self.sample = collection.new_sample()
            self._smiles_var = tk.StringVar(value="")
        else:
            self.sample = sample
            self._smiles_var = tk.StringVar(value=sample.molecule['cano_smiles'])
        self._smiles_var.trace_add("write", self._smiles_change)
        self._origen_smiles = self._smiles_var.get()

    def render(self, container: ttk.Frame):
        super().render(container)

        left_frame = ScrollableFrame(container, relief=tk.SUNKEN, padding=5)
        left_frame.pack(fill=tk.BOTH, expand=True)  # weight allows resizing
        left_frame = left_frame.scrollable_frame
        if self._is_new:
            ttk.Label(left_frame, text='New Sample', font=ConfigManager.header_font, justify='right').pack(fill=tk.X)
        else:
            ttk.Label(left_frame, text=f'Edit Sample {self.sample.short_label}', font=ConfigManager.header_font, justify='right').pack(fill=tk.X)

        entry_row = tk.Frame(left_frame)
        entry_row.pack(fill=tk.BOTH, pady=2)

        # Label (Fixed width)
        label = ttk.Label(entry_row, text="Sample SMILES:")
        label.grid(row=0, column=0, padx=(0, 5), sticky="w")
        entry_row.grid_columnconfigure(1, weight=1)# Left-aligned

        # Entry (Expands)
        entry = ttk.Entry(entry_row, textvariable=self._smiles_var)
        entry.grid(row=0, column=1, sticky="ew", padx=(0, 5))


        self._save_btn = ttk.Button(left_frame, style="Save.TButton", text="Save",
                              command=lambda *x: self._save())
        self._save_btn.pack()
        self._check_change()

    def _save(self):
        self.sample.molecule = self.instance.molecule().create_molecule_by_smiles(self._smiles_var.get())
        lg = LoadingGIF(self._page_manager.root)
        lg.add_label(self._save_btn)
        lg.start()
        def stop(success):
            lg.stop()
            if success:
                self._origen_smiles = self._smiles_var.get()
                self._check_change()
                self._page_manager.go_back()
                messagebox.showinfo("Saving Success", "The Sample was saved.")
            else:
                messagebox.showerror("Saving failed", "The Sample was not saved.")

        def run():
            try:
                self.sample.save()
                self._page_manager.root.after(1000, stop, True)
            except RequestException as e:
                self._page_manager.root.after(1000, stop, False)



        threading.Thread(target=run).start()


    def _smiles_change(self, *args):
        self._check_change()

    def _check_change(self):
        if self._origen_smiles == self._smiles_var.get():
            self._save_btn.state(["disabled"])  # Disable the button.
        else:
            self._save_btn.state(["!disabled"])


import json  # Import JSON module to read and write JSON files
import random
import sys  # Import sys to access command-line arguments

# Define the number of variations to generate
NUMBER_OF_VARIATIONS = 11

print(f'Generating {NUMBER_OF_VARIATIONS} random  variables')

# Read the input JSON file passed as the first command-line argument
print(f'Reading {sys.argv[1]}')
with open(sys.argv[1]) as jsonfile:
    values = json.loads(jsonfile.read())  # Load JSON content into a dictionary

# Calculate context-related values
print(f'Calculating context')
row_count = len(values)  # Total number of keys (variables) in the JSON data
number_of_options = len(values[list(values.keys())[0]]) - 1  # Number of options per variable (excluding unit)
print('Number of variables: ' + str(row_count))
print('Number of values per row: ' + str(number_of_options))
# Initialize results dictionary with VARIABLE and UNIT lists
results = {'VARIABLE': [], 'UNIT': []}

# Populate VARIABLE and UNIT lists with keys and corresponding unit values
for r_i, r_v in values.items():
    results['VARIABLE'].append(r_i)  # Store the variable name (key)
    results['UNIT'].append(r_v[0])  # Store the unit (first element of the value list)



# Generate variations and store them in the results dictionary
print(f'Preparing Variations')
for i in range(NUMBER_OF_VARIATIONS):
    results[f'Variation.{i}'] = []  # Initialize a new list for each variation
    for r_i, r_v in enumerate(values.values()):
        # Select a variation value based on computed index
        results[f'Variation.{i}'].append(random.choice(r_v[1:]))

# Write the results to the output JSON file
print(f'Writing {sys.argv[2]}')
with open(sys.argv[2], 'w+') as json_file:
    json_file.write(json.dumps(results, indent=4))  # Convert dictionary to JSON and write to file



import os
import re
import threading
import tkinter as tk
from datetime import datetime
from pathlib import Path
from tkinter import filedialog, ttk, messagebox

from chemotion_api import Reaction

from ChemDoE.config import ConfigManager
from ChemDoE.icons import IconManager
from ChemDoE.utils.dd_manager import DragManagerWidget
from ChemDoE.utils.pages import ToolBarPage, ScrollableFrame


class AnalysesUploader(ToolBarPage):
    def __init__(self, reaction: Reaction):
        super().__init__()
        self._container = []
        self.reaction = reaction
        self._folder_btn = None
        self._folder_path = None
        self._has_sub_dir = False
        self._files = []
        self._extensions = dict()
        self._is_tar = tk.BooleanVar(value=False)

    def select_directory(self):
        _folder_path = filedialog.askdirectory()
        if _folder_path:  # If a folder is selected
            self._folder_btn.config(text=f"Selected: {_folder_path}")
            self._folder_btn.config(width=len(_folder_path) + 10)
            self._folder_path = Path(_folder_path)
            self._check_sub_dirs()
            self._read_file_names()

    def render(self, container: ttk.Frame):
        super().render(container)
        mc = ttk.Frame(container, style='Background.TFrame')
        mc.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        scf = ScrollableFrame(mc, horizontal=True, style='Background.TFrame')
        for i in range(4):
            self._container.append(ttk.Frame(scf.scrollable_frame, style='Step.TFrame'))

        scf.pack(fill="both", expand=True)
        self.render_1()

    def render_1(self):
        container, clean = self._clean_container(0)
        ttk.Label(container, text="1) Select a folder", style='Header.TLabel').pack(padx=5, pady=5, anchor=tk.W)
        text = ("Please select a folder in which the analysis files are located. "
                "Please note that the files can also be located in subfolders of the selected "
                "folder. In order for the automatic assignment to work, it is necessary for the "
                "naming convention to be followed and for the file names to begin with the variation names.")
        ttk.Label(container, text=text, style='Para.TLabel', wraplength=500, justify="left").pack(padx=5, pady=5,
                                                                                                  anchor=tk.W)
        self._folder_btn = ttk.Button(container, text="Select Folder", command=self.select_directory)
        self._folder_btn.pack(padx=5, pady=5, anchor=tk.W)
        clean()

    def render_2(self):
        container, clean = self._clean_container(1)
        if not self._has_sub_dir:
            ttk.Label(container, text="2) Can be skipped ", style='Header.TLabel').pack(padx=5, pady=10, anchor=tk.W)
            text = "Step 2 is only needed if the selected directory has subdirectories"
            l = ttk.Label(container, text=text, style='Para.TLabel', wraplength=500, justify="left")
            l.pack(padx=5, pady=5, anchor=tk.W)
        else:
            ttk.Label(container, text="2) How to handle subdirectories", style='Header.TLabel').pack(padx=5, pady=10,
                                                                                                     anchor=tk.W)
            text = "Select every file in the subdirectory as single file or bundle them into a Tar archive"
            l = ttk.Label(container, text=text, style='Para.TLabel', wraplength=500, justify="left")
            l.pack(padx=5, pady=5, anchor=tk.W)

            c2 = ttk.Checkbutton(container, text='As Tar archive', onvalue=True, offvalue=False, variable=self._is_tar,
                                 command=self._read_file_names)
            c2.pack(padx=5, pady=5, anchor=tk.W)
        clean()

    def render_3(self):
        container, clean = self._clean_container(2)
        ttk.Label(container, text="3) Filter File extension", style='Header.TLabel').pack(padx=5, pady=10, anchor=tk.W)
        text = "Please select the relevant file extension to upload as an analysis."
        l = ttk.Label(container, text=text, style='Para.TLabel', wraplength=500, justify="left")
        l.pack(padx=5, pady=5, anchor=tk.W)

        ext_frame = ttk.Frame(container)
        ext_frame.pack(padx=5, pady=5, anchor=tk.W)

        for ext, ext_values in self._extensions.items():
            c2 = ttk.Checkbutton(ext_frame, text=ext, onvalue=True, offvalue=False, variable=ext_values,
                                 command=self._update_ext_filter)
            c2.pack(padx=5, pady=3, anchor=tk.W)
        clean()

    def render_4(self):
        container, clean = self._clean_container(3)
        ttk.Label(container, text="4) Map Files", style='Header.TLabel').pack(padx=5, pady=10, anchor=tk.W)
        text = "Please select the relevant file extension to upload as an analysis."
        l = ttk.Label(container, text=text, style='Para.TLabel', wraplength=500, justify="left")
        l.pack(padx=5, pady=5, anchor=tk.W)

        main_frame = ttk.Frame(container, style='Move.TFrame')
        column = 0
        cols = []
        dm = DragManagerWidget(self.page_manager.root)
        for name, files in self._file_mapping.items():
            row = 0
            header = ttk.Label(main_frame, text=name, style='TableHeader.TLabel', anchor='center')
            header.grid(row=row, column=column, sticky="ew")
            dm.on_drop_in_target(header, self._on_drop_in_target)
            for file in files['files']:
                row += 1
                col = ttk.Label(main_frame, text=file, style='TableItem.TLabel')
                col.grid(row=row, column=column, sticky="nsew")
                cols.append(col)
            column += 1
        dm.set_elements(cols)

        main_frame.pack(padx=5, pady=5, anchor=tk.W)
        ttk.Button(container, text="Transfer analyses files", image=IconManager().CHEMOTION, compound="left",
                   command=self._prepare_transfer).pack(anchor="w")
        clean()

    def _prepare_transfer(self):
        win = tk.Toplevel(self.page_manager.root, background="#ffffff" )

        win.title("Transfer")
        win.transient(self.page_manager.root)
        win.wm_overrideredirect(True)
        win.attributes('-topmost', 'true')
        win.resizable(False, False)
        win_width = 500
        win_height = 200
        x = self.page_manager.root.winfo_x() + self.page_manager.root.winfo_width() // 2 - win_width // 2
        y = self.page_manager.root.winfo_y() + self.page_manager.root.winfo_height() // 2 - win_height // 2

        win.geometry(f"{win_width}x{win_height}+{x}+{y}")


        label = ttk.Label(win, wraplength=15*30, text="Are you sure you want tp transfer the analyses files", style='Header.TLabel')
        progress_bar = ttk.Progressbar(win, length=300, mode="determinate")
        button = None

        btn_row = ttk.Frame(win)
        button = ttk.Button(btn_row, text="Transfer", command=lambda: self._transfer_analyses(win ,label, progress_bar, button, close_button))
        close_button = ttk.Button(btn_row, text="Close", command=win.destroy)
        label.pack(padx=5, pady=5)
        progress_bar.pack(padx=5, pady=5)
        btn_row.pack(padx=5, pady=5)
        button.pack(padx=5, pady=5, side=tk.LEFT)
        close_button.pack(padx=5, pady=5, side=tk.LEFT)

    def _transfer_analyses(self, win, label, progress_bar, button, close_button):
        label.config(text="Start transferring analyses files")
        button.pack_forget()
        close_button.pack_forget()
        progress_bar.pack(pady=20)
        progress_bar["value"] = 0
        texts = ['Uploading Files', 'Files Uploaded', 'linking Variations', 'Saved reaction. Analyses transferred.']
        step = 100 // len(texts)


        def update_bar(i):
            progress_bar["value"] = step * (i + 1)
            label.config(text=texts[i])
            if i + 1 == len(texts):
                close_button.pack(padx=5, pady=5)

        def upload():
            ana = self.reaction.analyses
            _current_anas = [a.id for a in ana]
            today = datetime.now().isoformat()

            for name, files in self._file_mapping.items():
                if name == 'Ignore':
                    continue
                for file in files['files']:
                    analyses = ana.add_analyses(f'{file}')
                    analyses['content'] = {
                        'ops': [{'attributes': {'bold': True}, 'insert': 'Variation: '}, {'insert': f'{name}\n'},
                                {'attributes': {'bold': True}, 'insert': 'Datime: '}, {'insert': f'{today}\n'},
                                {'attributes': {'bold': True}, 'insert': 'File: '}, {'insert': f'{file}\n'}]}
                    fp = os.path.join(self._folder_path, file)
                    analyses.add_dataset(str(fp))
            self.page_manager.root.after(0, update_bar, 0)
            self.reaction.save(True)
            self.page_manager.root.after(0, update_bar, 1)
            for ana in self.reaction.analyses:
                if ana.id not in _current_anas:
                    name = ana['content']['ops'][1]['insert'][:-1]
                    if self._file_mapping[name]['variation'] is not None:
                        v = next((s for s in self.reaction.variations if
                                  s.id == self._file_mapping[name]['variation']), None)
                        if v is not None:
                            v.link_analyses(ana)
            self.page_manager.root.after(0, update_bar, 2)
            self.reaction.save(True)
            self.page_manager.root.after(0, update_bar, 3)

        threading.Thread(target=upload).start()

    def _on_drop_in_target(self, src: tk.Widget, target: tk.Widget) -> bool:
        file_name = src.cget("text")
        target_name = target.cget("text")
        for name, files in self._file_mapping.items():
            if file_name in files['files']:
                files['files'].remove(file_name)
        self._file_mapping[target_name]['files'].append(file_name)
        column = target.grid_info()['column']
        src.grid_forget()
        src.grid(row=len(self._file_mapping[target_name]['files']), column=column, sticky="nsew")
        return False

    def set_style(self, style: ttk.Style):
        super().set_style(style)
        style.configure('Header.TLabel', font=ConfigManager.header_font)
        style.configure('Para.TLabel', font=ConfigManager.normal_font)
        style.configure('Background.TFrame', background="white")
        style.configure('Step.TFrame', borderwidth=2, relief=tk.SUNKEN)
        style.configure('Move.TFrame', borderwidth=2, relief=tk.SUNKEN)
        style.configure('TableHeader.TLabel', background="#d2f7ff", font=ConfigManager.header_font, borderwidth=1,
                        relief='solid', padding=10)
        style.configure('TableItem.TLabel', borderwidth=2, relief=tk.SOLID, padding=10, background='white')

    def _update_ext_filter(self):
        self._file_mapping = {}
        v_name_reg = re.compile(pattern=r'ChemDoE: <<(.+)>>')

        def find_v_names(text):
            match = re.search(v_name_reg, text)
            if match:
                return match.group(1)
            return None

        v_names = []
        for v in self.reaction.variations:
            name = find_v_names(v.notes)
            if name is not None:
                v_names.append(name)
                self._file_mapping[name] = {'variation': v.id, 'files': []}
        self._file_mapping['Reaction'] = {'variation': None, 'files': []}
        self._file_mapping['Ignore'] = {'variation': None, 'files': []}
        v_names.sort(key=lambda x: len(x), reverse=True)
        extensions = [x for x, var in self._extensions.items() if var.get()]
        for f in self._files:
            if any(f.endswith(ext) for ext in extensions):
                added = False
                for name in v_names:
                    if os.path.basename(f).startswith(name):
                        self._file_mapping[name]['files'].append(f)
                        added = True
                        break
                if not added:
                    self._file_mapping['Reaction']['files'].append(f)

        self.render_4()

    def _read_file_names(self):
        if self._is_tar.get():
            self._read_file_names_flat()
        else:
            self._read_file_names_deep()
            if len(self._files) > 50:
                self._is_tar.set(True)
                messagebox.showerror("Error",
                                     f"Too many files were selected. {len(self._files)} are too many files")
                self._read_file_names()
                return

        ext = {}
        for f in self._files:
            filename, file_extension = os.path.splitext(f)
            _filename, file_extension_2 = os.path.splitext(filename)
            if file_extension_2 == '.tar':
                file_extension = file_extension_2 + file_extension
            ext[file_extension] = self._extensions.get(file_extension, tk.BooleanVar(value=True))

        self._extensions = ext
        self.render_2()
        self.render_3()
        self._update_ext_filter()

    def _read_file_names_flat(self):
        self._files = []
        for f in self._folder_path.iterdir():
            if f.is_file():
                self._files.append(f.name)
            elif f.is_dir():
                self._files.append(f.name + '.tar.gz')

    def _read_file_names_deep(self):
        self._files = []
        for root, dir, files in self._folder_path.walk():
            root = root.relative_to(self._folder_path)
            for file in files:
                self._files.append(str(root / file))

    def _check_sub_dirs(self):
        self._has_sub_dir = False
        for f in self._folder_path.iterdir():
            if f.is_dir():
                self._has_sub_dir = True
                return

    def _clean_container(self, idx):
        if len(self._container) > idx + 1:
            for c in self._container[idx + 1:]:
                for widget in c.winfo_children():
                    widget.destroy()
                    del widget
                c.pack_forget()
        ws = [w for w in self._container[idx].winfo_children()]

        def clean():
            for w in ws:
                w.destroy()

        self._container[idx].pack(fill="x", padx=10, pady=10)
        return (self._container[idx], clean)



from tkinter import ttk, messagebox
import tkinter as tk
from urllib.parse import urlparse

from chemotion_api import Instance
from requests import RequestException

from ChemDoE.utils.keyboard_shortcuts import add_input_shortcuts
from ChemDoE.utils.page_manager import Page
from ChemDoE.start_view import StartPage

from ChemDoE.config import ConfigManager


class LoginManager(Page):
    def __init__(self):
        super().__init__()
        self.check_var = tk.IntVar(value=int(ConfigManager.get("Last", "remember")))
        self.host_var = tk.StringVar(value=ConfigManager.get("Last", "host"))
        self.host_var.trace_add("write", self._validate_input)
        self.check_label = self.username_entry = self.username_entry = None

    def _check_token(self):
        host = ConfigManager.get("Chemotion", "host")
        token = ConfigManager.get("Chemotion", "token")
        username = ConfigManager.get("Chemotion", "user")
        if host and token:
            try:
                instance = Instance(host).login_token(token).test_connection()
                ConfigManager().chemotion = instance

                ConfigManager().set('Last', 'host', host, commit=False)
                ConfigManager().set('Last', 'user', username)
                self.page_manager.start_page(StartPage())
                return True
            except (RequestException, PermissionError, ConnectionError):
                ConfigManager().chemotion = None
                pass
        return False

    def _login(self, *args):
        host = self.host_var.get()
        instance = Instance(host)
        username = self.username_entry.get()
        password = self.password_entry.get()


        try:
            instance.login(username, password).test_connection()
            ConfigManager().chemotion = instance
            if self.check_var.get():
                ConfigManager().set("Chemotion", "host", host, commit=False)
                ConfigManager().set("Chemotion", "user", username, commit=False)
                ConfigManager().set("Chemotion", "token", instance.token, commit=False)
            else:
                ConfigManager().set("Chemotion", "host", '', commit=False)
                ConfigManager().set("Chemotion", "user", '', commit=False)
                ConfigManager().set("Chemotion", "token", '', commit=False)

            ConfigManager().set('Last', 'remember', str(self.check_var.get()), commit=False)
            ConfigManager().set('Last', 'User', username)
            self.page_manager.start_page(StartPage())
        except RequestException:
            ConfigManager().chemotion = None
            messagebox.showerror("Login Failed", f"Invalid HOST url: {host}")
        except PermissionError:
            ConfigManager().chemotion = None
            messagebox.showerror("Login Failed", "Invalid username or password")

    def _validate_input(self, *args):
        """Check if input is a valid email and update the checkmark label."""
        host = self.host_var.get()
        instance = Instance(host)
        try:
            urlparse(host)
            instance.test_connection()
            self.check_label.config(text="✅", foreground="green")
            ConfigManager().set('Last', 'Host', host)
        except:
            self.check_label.config(text="", foreground="green")

    def leave(self):
        self.page_manager.root.unbind("<Return>")

    def render(self, root: ttk.Frame):
        if self._check_token():
            return
        frame = ttk.Frame(root, padding=20, style="Inner.TFrame")
        frame.pack(expand=True)

        row = 0

        ttk.Label(frame, text="Chemotion host:").grid(row=row, column=0, pady=5, sticky="w")
        url_entry = ttk.Entry(frame, textvariable=self.host_var)
        url_entry.grid(row=row, column=1, pady=5)
        add_input_shortcuts(url_entry)
        self.check_label = ttk.Label(frame, text="", font=("Arial", 14))
        self.check_label.grid(row=row, column=2, pady=5)
        row += 1

        # Username label and entry
        ttk.Label(frame, text="Username:").grid(row=row, column=0, pady=5, sticky="w")
        self.username_entry = ttk.Entry(frame)
        self.username_entry.grid(row=row, column=1, pady=5)
        add_input_shortcuts(self.username_entry)
        row += 1

        # Password label and entry
        ttk.Label(frame, text="Password:").grid(row=row, column=0, pady=5, sticky="w")
        self.password_entry = ttk.Entry(frame, show="*")
        self.password_entry.grid(row=row, column=1, pady=5)
        add_input_shortcuts(self.password_entry)

        row += 1

        # Login button
        login_button = ttk.Button(frame, text="Login", command=self._login)
        self.page_manager.root.bind("<Return>", self._login)
        login_button.grid(row=row, column=0, columnspan=1, pady=20)

        row += 1

        checkbox = ttk.Checkbutton(frame, text="Remember Me", variable=self.check_var)
        checkbox.grid(row=row, column=0, pady=5)

    def set_style(self, style: ttk.Style):
        style.configure("Inner.TFrame", background="#ffffff")



import os
import shutil

import cairosvg
import requests
from bs4 import BeautifulSoup

# Bootstrap Icons URL
# FontAwesome metadata URL
FA_ICONS_URL = "https://raw.githubusercontent.com/FortAwesome/Font-Awesome/master/metadata/icons.json"
FA_BASE_URL = "https://raw.githubusercontent.com/FortAwesome/Font-Awesome/master/svgs/solid/"


def run_download(output_folder: str = "fontawesome-icons"):
    # Folder to save icons
    if os.path.exists(output_folder):
        shutil.rmtree(output_folder)

    os.makedirs(output_folder, exist_ok=True)

    # Get the Bootstrap Icons page
    # Fetch icon metadata
    response = requests.get(FA_ICONS_URL)
    if response.status_code != 200:
        print("Failed to fetch FontAwesome icons metadata.")
        exit(1)

    icons = response.json()

    # Download each icon
    for icon_name in icons.keys():
        icon_url = f"{FA_BASE_URL}{icon_name}.svg"
        icon_path = os.path.join(output_folder, f"{icon_name}.svg.png")

        icon_response = requests.get(icon_url)
        if icon_response.status_code == 200:
            with open(icon_path, "wb") as file:
                png_data = cairosvg.svg2png(bytestring=icon_response.content, scale=2.0)
                file.write(png_data)
            print(f"Downloaded: {icon_name}.svg")
        else:
            print(f"Failed to download: {icon_name}.svg")

    print("Download complete.")


if __name__ == "__main__":
    run_download()


def clone_widget(widget, master=None):
    """
    Create a cloned version o a widget

    Parameters
    ----------
    widget : tkinter widget
        tkinter widget that shall be cloned.
    master : tkinter widget, optional
        Master widget onto which cloned widget shall be placed. If None, same master of input widget will be used. The
        default is None.

    Returns
    -------
    cloned : tkinter widget
        Clone of input widget onto master widget.

    """
    # Get main info
    parent = master if master else widget.master
    cls = widget.__class__

    # Clone the widget configuration
    cfg = {key: widget.cget(key) for key in widget.configure()}
    cloned = cls(parent, **cfg)

    # Clone the widget's children
    for child in widget.winfo_children():
        child_cloned = clone_widget(child, master=cloned)
        if child.grid_info():
            grid_info = {k: v for k, v in child.grid_info().items() if k not in {'in'}}
            child_cloned.grid(**grid_info)
        elif child.place_info():
            place_info = {k: v for k, v in child.place_info().items() if k not in {'in'}}
            child_cloned.place(**place_info)
        else:
            pack_info = {k: v for k, v in child.pack_info().items() if k not in {'in'}}
            child_cloned.pack(**pack_info)

    return cloned


import tkinter as tk


def add_input_shortcuts(entry):
    def select_all(event):
        event.widget.select_range(0, tk.END)  # Select all text
        event.widget.icursor(tk.END)  # Move cursor to the end
        return "break"

    entry.bind("<Control-a>", select_all)


def add_placeholder(entry, placeholder):
    entry.bind("<FocusIn>", lambda x: _remove_placeholder(entry, placeholder))
    entry.bind("<FocusOut>", lambda x: _add_placeholder(entry, placeholder))
    _add_placeholder(entry, placeholder)


def _add_placeholder(entry, placeholder):
    """Show placeholder text when Entry is empty and loses focus."""
    if not entry.get():
        entry.insert(0, placeholder)
        entry.config(foreground="gray")  # Placeholder color


def _remove_placeholder(entry, placeholder):
    """Remove placeholder text when user starts typing."""
    if entry.get() == placeholder:
        entry.delete(0, tk.END)
        entry.config(foreground="black")  # Normal text color



from chemotion_api import Instance

from ChemDoE.element_tree_page import ElementTreePage
from ChemDoE.new_reaction import NewReaction


class StartPage(ElementTreePage):
    def __init__(self):
        super().__init__('Reaction')
        self._new_reaction = None


    def create_new(self, collection):
        self._new_reaction = NewReaction(collection)
        self.page_manager.set_page(self._new_reaction)


    def select_element(self, collection, reaction):
        self._new_reaction = NewReaction(collection, reaction)
        self.page_manager.set_page(self._new_reaction)


import threading
import webbrowser
from tkinter import ttk
import tkinter as tk
from tkinter.ttk import Combobox
from typing import Optional, Callable

from ChemDoE.config import ConfigManager
from ChemDoE.icons import IconManager
from ChemDoE.utils.page_manager import Page


class EasyOptionMenu(tk.OptionMenu):
    def __init__(self, master, *values: str, **kwargs):
        self.var = tk.StringVar()
        self.silent = False
        super().__init__(master, self.var, "", **kwargs)
        self.values = list(values)

        self._on_select = lambda _v, _i: None
        self._current = -1
        self.var.trace_add("write", lambda _a, _b, _c: self.on_select(self.var.get(), self.current_idx))

    @property
    def on_select(self) -> Callable[[str, int], None]:
        if self.silent:
            return lambda _v, _i: None
        return self._on_select

    @on_select.setter
    def on_select(self, value: Callable[[str, int], None]):
        self._on_select = value

    @property
    def values(self) -> list[str]:
        return self._values

    @property
    def current_idx(self):
        return self._current

    @current_idx.setter
    def current_idx(self, i):
        self._set_var_on_select(self.values[i], i)

    @values.setter
    def values(self, new_options: list[str]):
        menu = self["menu"]
        menu.delete(0, "end")  # Clear existing options
        for idx, option in enumerate(new_options):
            menu.add_command(label=option,
                             command=lambda v=option, i=idx: self._set_var_on_select(v, i))  # Update options
        self.var.set('')
        self._current = -1
        self._values = new_options  # Set a default value

    def insert_separator(self, idx: int):
        self["menu"].insert_separator(idx)

    def _set_var_on_select(self, val, i):
        self._current = i
        self.var.set(val)

    def get(self):
        return self.var.get()

    def current(self, i: int = None):
        if i is not None:
            self.current_idx = i
        return self.current_idx

    def set(self, val: str):
        idx = self.values.index(val)
        self._set_var_on_select(val, idx)


class ScrollableFrame(ttk.Frame):
    """
    A scrollable frame that can be embedded in other widgets.
    It creates a canvas with a vertical scrollbar and places a frame inside the canvas.
    """

    def __init__(self, container, horizontal=False, *args, **kwargs):
        super().__init__(container, *args, **kwargs)

        # Create a canvas widget inside the frame
        self._yscroll_disabled = True
        self.canvas = tk.Canvas(self)
        self.canvas.grid(row=0, column=0, sticky="nsew")

        # Create a vertical scrollbar linked to the canvas
        self.scrollbar = ttk.Scrollbar(self, orient="vertical", command=self.canvas.yview)
        self.scrollbar.grid(row=0, column=1, sticky="ns")
        # Configure the canvas to use the scrollbar
        self.canvas.configure(yscrollcommand=self.scrollbar.set)
        if horizontal:
            self.scrollbar_x = ttk.Scrollbar(self, orient="horizontal", command=self.canvas.xview)
            # Configure the canvas to use the scrollbar
            self.canvas.configure(xscrollcommand=self.scrollbar_x.set)
            self.scrollbar_x.grid(row=1, column=0, sticky="ew")
        else:
            self.scrollbar_x = None

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        # Create a frame inside the canvas which will hold the content
        self.scrollable_frame = ttk.Frame(self.canvas, style='Background.TFrame')

        # Create a window inside the canvas to hold the inner frame
        self.window_item = self.canvas.create_window(0, 0, window=self.scrollable_frame, anchor="nw")

        # Update the scrollregion of the canvas whenever the size of the inner frame changes
        self.canvas.bind(
            "<Configure>",
            self._on_canvas_configure
        )

        self.bind('<Enter>', self._bound_to_mousewheel)
        self.bind('<Leave>', self._unbound_to_mousewheel)

    def destroy(self):
        self._unbound_to_mousewheel(None)
        self.canvas.destroy()
        super().destroy()

    def _bound_to_mousewheel(self, event):
        self.canvas.bind_all("<MouseWheel>", self._on_mouse_scroll)
        self.canvas.bind_all("<Button-4>", self._on_mouse_scroll)
        self.canvas.bind_all("<Button-5>", self._on_mouse_scroll)

    def _unbound_to_mousewheel(self, event):
        self.canvas.unbind_all("<MouseWheel>")
        self.canvas.unbind_all("<Button-4>")
        self.canvas.unbind_all("<Button-5>")

    def update_sc_view(self):
        self._on_canvas_configure(None)

    def _on_canvas_configure(self, event):
        # Set the inner frame's width to match the canvas's width.
        bbox = self.canvas.bbox("all")
        self.canvas.configure(scrollregion=bbox)
        self.update_scroll_visibility(bbox)
        self.update_frame_width()

    def update_frame_width(self):
        """Adjust frame width if canvas is wider than the content"""
        canvas_width = self.canvas.winfo_width()
        frame_width = self.scrollable_frame.winfo_reqwidth()

        if canvas_width > frame_width:
            self.canvas.itemconfig(self.window_item, width=canvas_width)
        else:
            self.canvas.itemconfig(self.window_item, width=frame_width)

    def update_scroll_visibility(self, bbox):
        """Show scrollbars only if needed"""
        canvas_width = self.canvas.winfo_width()
        canvas_height = self.canvas.winfo_height()

        if bbox:
            frame_width = bbox[2] - bbox[0]
            frame_height = bbox[3] - bbox[1]

            # Show/hide vertical scrollbar
            if frame_height > canvas_height:
                self._yscroll_disabled = False
                self.scrollbar.grid(row=0, column=1, sticky="ns")
                self.canvas.configure(yscrollcommand=self.scrollbar.set)
            else:
                self._yscroll_disabled = True
                self.scrollbar.grid_remove()
                self.canvas.configure(yscrollcommand="")

            # Show/hide horizontal scrollbar
            if self.scrollbar_x and frame_width > canvas_width:
                self.scrollbar_x.grid(row=1, column=0, sticky="ew")
                self.canvas.configure(xscrollcommand=self.scrollbar_x.set)
            elif self.scrollbar_x:
                self.scrollbar_x.grid_remove()
                self.canvas.configure(xscrollcommand="")

    def _on_mouse_scroll(self, event):
        if self._yscroll_disabled:
            return
        """Enable scrolling with mouse wheel"""
        if event.num == 4:  # Scroll Up (Linux)
            self.canvas.yview_scroll(-1, "units")
        elif event.num == 5:  # Scroll Down (Linux)
            self.canvas.yview_scroll(1, "units")
        elif event.delta:  # Windows & macOS
            self.canvas.yview_scroll(-1 * (event.delta // 120), "units")


class ListRow(tk.Frame):
    def __init__(self, parent, icon, title, subtitle, delete_callback):
        super().__init__(parent, bg="white", padx=5, pady=5)

        # Icon Label (Using Emoji as Placeholder)
        self.icon_label = tk.Label(self, image=icon, font=("Arial", 16), bg="white")
        self.icon_label.pack(side="left", padx=5)

        # Text Frame (Title + Subtitle)
        text_frame = tk.Frame(self, bg="white")
        text_frame.pack(side="left", fill="both", expand=True)

        self.title_label = tk.Label(text_frame, text=title, font=("Arial", 12, "bold"), bg="white")
        self.title_label.pack(anchor="w")

        self.subtitle_label = tk.Label(text_frame, text=subtitle, font=("Arial", 10), fg="gray", bg="white")
        self.subtitle_label.pack(anchor="w")

        # Delete Button
        self.delete_button = ttk.Button(self, image=IconManager().TRASH, width=3, command=lambda: delete_callback(self))
        self.delete_button.pack(side="right", padx=5)


class ToolBarPage(Page):
    fav_dropdown: Combobox

    def __init__(self):
        super().__init__()
        self._fav_reactions = None
        self.instance = ConfigManager().chemotion
        self.toolbar: Optional[ttk.Frame] = None

    def _logout(self):
        from ChemDoE.login_manager import LoginManager
        ConfigManager().logout()
        self.page_manager.start_page(LoginManager())

    def render(self, container: ttk.Frame):
        # Create a toolbar frame at the top
        toolbar = ttk.Frame(container, padding=0)
        toolbar.pack(side="top", fill="x")

        if self.page_manager.can_go_back():
            back_btn = ttk.Button(toolbar, style="NAV.Active.TButton", image=IconManager().BACK_ICON,
                                  command=lambda *x: self.page_manager.go_back())
        else:
            back_btn = ttk.Button(toolbar, style="NAV.NotActive.TButton", image=IconManager().BACK_ICON)

        if self.page_manager.can_go_forward():
            forward_btn = ttk.Button(toolbar, style="NAV.Active.TButton", image=IconManager().FORWARD_ICON,
                                     command=lambda *x: self.page_manager.go_forward())
        else:
            forward_btn = ttk.Button(toolbar, style="NAV.NotActive.TButton", image=IconManager().FORWARD_ICON)
        back_btn.image = IconManager().BACK_ICON
        forward_btn.image = IconManager().FORWARD_ICON
        back_btn.pack(side="left", padx=5)
        forward_btn.pack(side="left", padx=5)

        self._fav_reactions = []
        # Create a Combobox with the list of options
        self.fav_dropdown = ttk.Combobox(toolbar, values=[], state="readonly")
        self.fav_dropdown.pack(side="left", padx=5, pady=0)
        self.fav_dropdown.bind("<<ComboboxSelected>>", self._on_fav_select)
        url = f'{self.instance.host_url}/mydb'
        if hasattr(self, 'reaction') and self.reaction.id:
            url += f'/collection/{self.instance.get_root_collection().id}/reaction/{self.reaction.id}'

        ttk.Button(toolbar, image=IconManager().CHEMOTION, command=lambda: webbrowser.open(url), style="Info.TButton").pack(side="left", padx=5)
        ttk.Button(toolbar, image=IconManager().INFO, command=lambda: webbrowser.open('https://chemdoe.readthedocs.io/en/latest'), style="Info.TButton").pack(side="left", padx=5)

        self.toolbar = ttk.Frame(toolbar, padding=0)
        self.toolbar.pack(side="left", padx=5)

        logout_button = ttk.Button(toolbar, text="Logout", style="Logout.TButton", command=self._logout)
        logout_button.pack(side="right", padx=5)
        self.update_fav_dropdown()

    def update_fav_dropdown(self):
        def load():
            self._fav_reactions = [(-1, 'Favorites')] + ConfigManager().favorites_with_names
            self._page_manager.root.after(0, done_load)

        def done_load():
            self.fav_dropdown.config(values=[x[1] for x in self._fav_reactions])
            if hasattr(self, 'reaction'):
                idx = next((i for i, x in enumerate(self._fav_reactions) if x[0] == self.reaction.id), 0)
            else:
                idx = 0
            self.fav_dropdown.current(idx)

        t = threading.Thread(target=load)
        t.daemon = True
        t.start()

    def _on_fav_select(self, event):
        idx = self.fav_dropdown.current()
        id = self._fav_reactions[idx][0]
        if id > 0:
            from ChemDoE.new_reaction import NewReaction
            reaction = ConfigManager().chemotion.get_reaction(id)
            try:
                collection_id = reaction.json_data['tag']['taggable_data']['collection_labels'][0]["id"]
                col = ConfigManager().chemotion.get_root_collection().find(id=collection_id)[0]
            except (KeyError, IndexError):
                col = ConfigManager().chemotion.get_root_collection()
            self._page_manager.set_page(NewReaction(col, reaction))

    def set_style(self, style: ttk.Style):

        button_default = dict(
            padding=2,  # Left, Top, Right, Bottom
            borderwidth=0,  # Removes all borders
            relief="flat"
        )

        style.configure("Logout.TButton", background="#f54260",
                        foreground="white", **button_default)
        style.configure("NAV.Active.TButton", **button_default)
        style.configure("NAV.NotActive.TButton", background="#756b6d", foreground="lightgray", **button_default)
        style.map("NAV.NotActive.TButton", background=[("active", "#756b6d")], foreground=[("active", "lightgray")])

        style.configure("Outer.TFrame", background="white")
        style.configure("Info.TButton", background="#c2e7ff")
        style.map("Info.TButton", background=[("active", "#d2f7ff")], foreground=[("active", "white")])



import csv
import json
from pathlib import Path
from typing import LiteralString


def read_results(file_path: str | LiteralString | Path) -> dict:
    fp = Path(file_path)
    if fp.suffix == ".json":
        return read_json(fp)
    if fp.suffix == ".csv":
        return read_csv(fp)
    raise ValueError("Unsupported file type")


def read_json(file_path: str | LiteralString | Path) -> dict:
    with open(file_path, 'r', encoding='utf-8') as json_file:
        results = json.loads(json_file.read())
        return results


def read_csv(file_path: str | LiteralString | Path) -> dict:
    with open(file_path, newline='', encoding='utf-8') as csvfile:
        reader = csv.reader(csvfile)
        headers = next(reader, None)  # Get headers from the first row
        results = {}
        if headers:
            for h in list(headers):
                results[h] = []
            for row in reader:
                for i, v in enumerate(row):
                    results[headers[i]].append(v)

        return results



import os
import sys
import threading
import uuid
import webbrowser
from pathlib import Path
from tkinter import ttk
import tkinter as tk
from tkinter import filedialog
import tkinter.font as tkFont
from typing import Optional, Callable

from chemotion_api import Reaction

from ChemDoE.config import ConfigManager
from ChemDoE.execute import ExecuteManager
from ChemDoE.icons import IconManager

from ChemDoE.utils.keyboard_shortcuts import add_input_shortcuts
from ChemDoE.utils.pages import ScrollableFrame, EasyOptionMenu


class ScriptOrganizer(tk.Frame):
    def __init__(self, page_manager, master, reaction: Reaction, **kwargs):
        super().__init__(master, **kwargs)
        self._selected_template = EasyOptionMenu(self)
        self._fill_dropdown()
        self._run_btn = ttk.Button(self, text="▶", style='Run.TButton', width=2, command=self.run)
        self._selected_template.pack(side=tk.LEFT)
        self._selected_template.on_select = self._on_select
        self._run_btn.pack(side=tk.LEFT, padx=5, pady=0)
        self._root = page_manager.root
        self._page_manager = page_manager
        self._values = {}
        self._reaction = reaction

    @property
    def values(self) -> dict:
        return self._values

    @values.setter
    def values(self, values: dict):
        self._values = values

    def run(self):
        script = self._scripts[self._selected_template.current()]
        e = ExecuteManager(self._root, self._reaction)

        def run():
            e.run(script, self._values)

        threading.Thread(target=run).start()

    def _fill_dropdown(self):
        self._scripts = ConfigManager().load_scripts()
        self._selected_template.values = [x['name'] for x in self._scripts] + ["🛠 Edit configurations", '📖 How to?']
        self._selected_template.insert_separator(len(self._selected_template.values) - 2)
        self._set_last_run()

    def _set_last_run(self):
        rt = ConfigManager().get('Last', 'run_template')
        idx = 0
        if rt is not None:
            idx = next((i for i, x in enumerate(self._scripts) if x['id'] == rt), 0)
        self._selected_template.current(idx)

    def _on_select(self, val, idx):
        if idx == len(self._scripts) + 1:
            webbrowser.open('https://chemdoe.readthedocs.io/en/latest/executing_scripts.html')
        if idx == len(self._scripts):
            top = ScriptManager(self._root, self._scripts)

            def manager_on_close():
                self._selected_template.silent = True
                self._fill_dropdown()
                if top.last_edited != '':
                    self._selected_template.set(top.last_edited)
                self._selected_template.silent = False
                top.destroy()

            top.protocol("WM_DELETE_WINDOW", manager_on_close)
        else:
            current_script = self._scripts[idx]
            ConfigManager().set('Last', 'run_template', current_script['id'])

    @staticmethod
    def set_style(style: ttk.Style):

        style.configure('Run.TButton', borderwidth=1, relief='solid', padding=(0, 0))
        style.configure('FileLoad.TButton', borderwidth=1, relief='solid', padding=2, font=ConfigManager.small_font)
        style.configure('Remove.TButton', background='#d9534f')

        style.map('Run.TButton',
                  foreground=[('!active', 'white'), ('pressed', 'white'), ('active', '#333333')],
                  background=[('!active', '#5cb85c'), ('pressed', '#5cb85c'), ('active', '#5cb85c')]
                  )


class ScriptForm(ScrollableFrame):
    default_file_text = "... select a file"

    def __init__(self, master, scripts, **kwargs):
        super().__init__(master, horizontal=True, **kwargs)

        self._interpreter_var = tk.StringVar()
        self._name_var = tk.StringVar(value="New Script")
        self._type_var = tk.StringVar(value="R")
        self._type_var.trace_add("write", self._change_type)
        self._input_type_var = tk.StringVar(value="CSV")
        self._output_type_var = tk.StringVar(value="CSV")
        self._file_path = None
        self._id = None
        self._scripts = scripts
        self._on_save: Callable[[dict], None] = lambda x: None

        frame = ttk.Frame(self.scrollable_frame, padding=(5), style="Inner.TFrame")
        frame.pack(expand=True)

        row = 0

        ttk.Label(frame, text="Name:").grid(row=row, column=0, pady=5, sticky="w")
        names_entry = ttk.Entry(frame, textvariable=self._name_var)
        names_entry.grid(row=row, column=1, pady=5, sticky="we", padx=5)
        add_input_shortcuts(names_entry)

        row += 1

        # Username label and entry
        ttk.Label(frame, text="Script type:").grid(row=row, column=0, pady=5, sticky="w")
        ttk.Combobox(frame, values=['R', 'Python'], textvariable=self._type_var, state="readonly").grid(row=row, column=1, pady=5, padx=5,
                                                                                      sticky="we")

        row += 1

        # Username label and entry
        ttk.Label(frame, text="Interpreter:").grid(row=row, column=0, pady=5, sticky="w")
        self._interpreter_select = ttk.Combobox(frame, values=[], textvariable=self._interpreter_var)
        self._interpreter_select.grid(row=row, column=1, pady=5, padx=5, sticky="we")

        row += 1

        # Username label and entry
        ttk.Label(frame, text="Script path:").grid(row=row, column=0, pady=5, sticky="w")
        input_frame = ttk.Frame(frame)
        self._file_input = ttk.Button(input_frame, text=self.default_file_text, command=self.open_file,
                                      style="FileLoad.TButton")
        input_frame.grid(row=row, column=1, sticky="we", pady=5)
        self._file_input.pack(side="left")
        self._copy_button = ttk.Button(input_frame, text='Open', command=self.copy_file_dir,
                                       style="FileLoad.TButton")
        self._copy_button.pack(side="left")
        self._copy_button.config(state=tk.DISABLED)
        row += 1

        # Username label and entry
        ttk.Label(frame, text="INPUT type:").grid(row=row, column=0, pady=5, sticky="w")
        ttk.Combobox(frame, values=['CSV', 'JSON'], textvariable=self._input_type_var, state="readonly").grid(row=row, column=1, pady=5,
                                                                                            padx=5, sticky="we")

        row += 1

        # Username label and entry
        ttk.Label(frame, text="OUTPUT type:").grid(row=row, column=0, pady=5, sticky="w")
        ttk.Combobox(frame, values=['CSV', 'JSON'], textvariable=self._output_type_var, state="readonly").grid(row=row, column=1, pady=5,
                                                                                             padx=5, sticky="we")

        # You can also add buttons, entries, etc. to the subwindow
        bt_frame = ttk.Frame(self.scrollable_frame)
        bt_frame.pack(pady=10)
        close_button = ttk.Button(bt_frame, text="Save", command=self.save)
        close_button.pack(side='left')
        self.rm_btn = ttk.Button(bt_frame, text="Remove", command=self.remove, style="Remove.TButton")

    def copy_file_dir(self):
        foldername = str(self._file_path)
        if sys.platform.startswith('linux'):
            os.system('xdg-open "%s"' % foldername)
        elif sys.platform.startswith('win32'):
            os.startfile(foldername)
        elif sys.platform.startswith('foldername'):
            os.system('open "%s"' % foldername)

    @property
    def on_save(self) -> Callable[[dict], None]:
        return self._on_save

    @on_save.setter
    def on_save(self, value: Callable[[dict], None]):
        self._on_save = value

    def set_value(self, as_dict: dict):
        self._as_dict = as_dict
        self._set_values(**as_dict)

    def _set_values(self, id, name, file_type, input, file, interpreter=None, output=None):

        if interpreter is None:
            interpreter = file_type.lower()
        if output is None:
            output = input
        self._set_file(file)
        if id is None or id.startswith("__default_"):
            self._id = None
            self.rm_btn.pack_forget()
        else:
            self.rm_btn.pack(side='left', padx=5)
            self._id = str(id)
        self._name_var.set(name)
        self._type_var.set(file_type)
        self._interpreter_var.set(interpreter)
        self._input_type_var.set(input)
        self._output_type_var.set(output)

    def to_dict(self, values=None):
        if values is None:
            values = {}

        if self._id is None:
            self._id = values['id'] = uuid.uuid4().__str__()
        else:
            values['id'] = self._id

        values['name'] = self._name_var.get()
        if values['name'].startswith('[Default] '):
            values['name'] = values['name'][len('[Default] '):]

        values['file_type'] = self._type_var.get()
        values['input'] = self._input_type_var.get()
        values['output'] = self._output_type_var.get()
        values['file'] = self._file_path.__str__()
        values['interpreter'] = self._interpreter_var.get()
        return values

    def save(self):
        if self._id is None:
            values = self.to_dict()
            self.set_value(values)
            self._scripts.append(values)
        else:
            values = self.to_dict(self._as_dict)
        ConfigManager().save_scripts(self._scripts)
        self._on_save(values)

    def remove(self):
        if self._id is None:
            return
        values = self.to_dict()
        self._scripts.remove(values)
        values['id'] = None
        self.set_value(values)
        ConfigManager().save_scripts(self._scripts)
        self._on_save(None)

    def open_file(self):
        # Open a file dialog and return the selected file's path

        top = self.winfo_toplevel()

        filename = filedialog.askopenfilename(
            title="Select a File",
            filetypes=(("Python files", "*.py"), ("R files", "*.R"))
        )

        top.lift()  # Bring it to the front
        top.attributes("-topmost", True)  # Keep it on top
        top.after(100, lambda: top.attributes("-topmost", False))

        if filename:
            self._set_file(filename)

            # You can now open and process the file as needed

    def _change_type(self, _a: str, _b: str, _c: str):
        if self._type_var.get() == 'Python':
            options = ConfigManager().python_interpreters + [self._type_var.get().lower()]
        elif self._type_var.get() == 'R':
            options = ConfigManager().r_interpreters + [self._type_var.get().lower()]
        else:
            options = []
        self._interpreter_select.config(values=options)
        self._interpreter_select.current(0)
        font = tkFont.Font(font=self._interpreter_select.cget("font"))
        max_width = max(font.measure(option) for option in options)  # Get max text width in pixels
        self._interpreter_select.config(width=max_width // font.measure("0") + 2)

    def _set_file(self, filename):
        self._file_path = Path(filename)
        bt_text = filename
        if self._file_path.suffix.lower() == '.py':
            self._type_var.set('Python')
        elif self._file_path.suffix.lower() == '.r':
            self._type_var.set('R')
        else:
            bt_text = self.default_file_text
            self._file_path = None
        self._file_input.config(text=bt_text, width=len(bt_text) + 5)
        if self._file_path is not None:
            self._copy_button.config(state=f'!{tk.DISABLED}')
        if self._file_path is None:
            self._copy_button.config(state=tk.DISABLED)


class ScriptManager(tk.Toplevel):

    def __init__(self, root, scripts):
        super().__init__(root)
        self.title("Run configurations manager")

        w, h = root.winfo_screenwidth(), root.winfo_screenheight()
        self.geometry("%dx%d" % (min(w - 100, 1000), min(h - 100, 1000)))

        self.transient(root)

        self._scripts = scripts
        self._sf: Optional[ScriptForm] = None
        self._last_edited: str = ''

        # Add some content to the subwindow
        label = ttk.Label(self, text="Edit Run configurations")
        label.pack(fill="x", pady=5, padx=5)

        paned_window = ttk.PanedWindow(self, orient=tk.HORIZONTAL)
        paned_window.pack(fill=tk.BOTH, expand=True)

        left_frame = ttk.Frame(paned_window, relief=tk.SUNKEN)
        paned_window.add(left_frame, weight=1)

        self._tree = ttk.Treeview(left_frame)
        self._tree.heading("#0", text="Run configurations", anchor="w")

        self._tree.insert("", "end", text=f'Add new', open=True,
                          image=IconManager().PLUS_ICON,
                          values=('_', 'Add_new'))

        self._python_node = self._tree.insert("", "end", text=f'Python', open=True,
                                              image=IconManager().PYTHON,
                                              values=('_', 'PYTHON'))

        self._r_node = self._tree.insert("", "end", text=f'R', open=True,
                                         image=IconManager().R,
                                         values=('_', 'R'))

        self._fill_scripts()

        self._tree.pack(fill=tk.BOTH)

        self._tree.bind("<ButtonRelease-1>", self._select_config)

        self._right_frame = ttk.Frame(paned_window, width=500, relief=tk.SUNKEN)
        paned_window.add(self._right_frame, weight=2)

    def _fill_scripts(self):
        for parent_node in [self._r_node, self._python_node]:
            for child in self._tree.get_children(parent_node):
                self._tree.delete(child)

        for s in self._scripts:
            if s['file_type'] == 'R':
                self._tree.insert(self._r_node, tk.END, text=s['name'], values=(s['id'], s['file_type']))
            elif s['file_type'] == 'Python':
                self._tree.insert(self._python_node, tk.END, text=s['name'], values=(s['id'], s['file_type']))

    def _on_save(self, as_dict: dict):
        self._fill_scripts()
        if as_dict:
            self._last_edited = as_dict.get('name', '')

    @property
    def last_edited(self) -> str:
        return self._last_edited

    def _select_config(self, event):
        """Handles single-click event."""
        selected_item = self._tree.focus()  # Get selected item
        item_text = self._tree.item(selected_item, "values")
        if len(item_text) < 2:
            return
        if item_text[1] == 'Add_new':
            if self._sf:
                self._sf.destroy()
            self._sf = ScriptForm(self._right_frame, self._scripts)
            self._sf.pack(fill=tk.BOTH, expand=True)
        elif item_text[0] != '_':
            if not self._sf:
                self._sf = ScriptForm(self._right_frame, self._scripts)
                self._sf.pack(fill=tk.BOTH, expand=True)

            self._sf.set_value(next((x for x in self._scripts if x['id'] == item_text[0])))

        else:
            return
        self._sf.on_save = self._on_save



import threading
from tkinter import ttk, messagebox
import tkinter as tk
from typing import Optional

from chemotion_api import Instance, Reaction, Sample
from chemotion_api.collection import Collection
from requests import RequestException

from ChemDoE.analyses_uploader import AnalysesUploader
from ChemDoE.config import ConfigManager
from ChemDoE.element_tree_page import ElementTreePage
from ChemDoE.icons import IconManager, LoadingGIF
from ChemDoE.new_sample import NewSample
from ChemDoE.utils.dd_manager import DragManagerTree
from ChemDoE.utils.pages import ListRow, ScrollableFrame
from ChemDoE.doe_manager import DoEPage


class NewReaction(ElementTreePage):
    def __init__(self, collection: Collection, reaction: Optional[Reaction] = None):
        super().__init__('Sample')
        self._collection = collection
        self._is_new = reaction is None
        self._dh: Optional[DragManagerTree] = None
        if self._is_new:
            self.reaction = collection.new_reaction()
            self._name_var = tk.StringVar(value="New Reaction")
        else:
            self.reaction = reaction
            self._name_var = tk.StringVar(value=reaction.name)
        self._name_var.trace_add("write", self._name_change)
        self._origen_data = self._prepare_compare_data()

    def render(self, container: ttk.Frame):
        super().render(container)
        if not self._is_new:
            self.reaction = self.instance.get_reaction(self.reaction.id)
        outher_frame = tk.Frame(self.paned_window, bg="")
        left_frame = ScrollableFrame(outher_frame, relief=tk.SUNKEN, padding=5)
        left_frame.pack(side=tk.LEFT, fill="both", expand=True)
        self.paned_window.add(outher_frame, weight=3)  # weight allows resizing
        left_frame = left_frame.scrollable_frame
        button_row = tk.Frame(left_frame, bg="white")
        self._save_btn = ttk.Button(button_row, style="Save.TButton", text="Save",
                              command=lambda *x: self._save())
        self._save_btn.pack(side=tk.LEFT)
        self._doe_btn = ttk.Button(button_row, style="Menu.TButton", text="Adjust DoE",
                              command=lambda *x: self._go_to_doe())

        self._doe_btn.pack(side=tk.LEFT)
        ana_btn = ttk.Button(button_row, style="Menu.TButton", text="Upload Analyses",
                              command=lambda *x: self._go_to_ana())
        ana_btn.pack(side=tk.LEFT)
        if not any(x.notes.startswith('ChemDoE: ') for x in self.reaction.variations):
            ana_btn.state(["disabled"])

        self._fav_btn = ttk.Button(button_row, style="Fav.TButton", text="Add ot Favorites",
                              command=lambda *x: self._toggel_favorites())
        self._fav_btn.pack(side=tk.RIGHT)
        button_row.pack(fill="x", expand=True)

        if self._is_new:
            ttk.Label(left_frame, text='New Reaction', font=ConfigManager.header_font, justify='right').pack(fill=tk.X)
        else:
            ttk.Label(left_frame, text=f'Edit Reaction {self.reaction.short_label}', font=ConfigManager.header_font, justify='right').pack(fill=tk.X)

        entry_row = tk.Frame(left_frame)
        entry_row.pack(fill="x", pady=2)

        # Label (Fixed width)
        label = ttk.Label(entry_row, text="Reaction name:")
        label.grid(row=0, column=0, padx=(0, 5), sticky="w")
        entry_row.grid_columnconfigure(1, weight=1)# Left-aligned

        # Entry (Expands)
        entry = ttk.Entry(entry_row, textvariable=self._name_var)
        entry.grid(row=0, column=1, sticky="ew", padx=(0, 5))


        self._dh = DragManagerTree(self.collection_tree, self.page_manager.root)
        self._dh.on_drag_start = self._drag_start

        self._render_material_input(left_frame, 'Starting Material', self.reaction.properties['starting_materials'])
        self._render_material_input(left_frame, 'Reactants', self.reaction.properties['reactants'])
        self._render_material_input(left_frame, 'Products', self.reaction.properties['products'])
        self._check_change()
        self._style_favorite()


    def _render_material_input(self, left_frame, title, elements):
        start_mat_row = tk.Frame(left_frame, borderwidth=1, relief=tk.RAISED, background="#ffffff")
        ttk.Label(start_mat_row, text=title).pack(fill=tk.X)
        list_frame = tk.Frame(start_mat_row)
        self._list_materials(elements, list_frame)
        list_frame.pack(fill="x", pady=(0, 5))
        drop_frame = tk.Frame(start_mat_row, background='#aaaaaa')
        drop_lable = ttk.Label(drop_frame, text='Drop New Material', anchor="center", style='Drop.TLabel')
        drop_lable.pack(expand=True, fill=tk.X, pady=5, padx=5)
        start_mat_row.pack(fill=tk.X, pady=5)
        drop_frame.pack(fill=tk.X, pady=5)
        self._dh.on_drop_in_target(drop_frame, lambda selection, target: self._on_drop(str(selection), list_frame, elements))

    def _list_materials(self, materials: list[Sample], frame):
        for idx, sample in enumerate(materials):
            self._list_sdd_materials(materials, idx, frame)


    def _list_sdd_materials(self, materials: list[Sample], idx: int, frame):
        sample = materials[idx]
        row = ListRow(frame, IconManager().SAMPLE_ICON, sample.short_label, sample.molecule['cano_smiles'], lambda *args: self._delete_row(sample, materials, row, *args))
        row.pack(fill="x", pady=2)

    def _delete_row(self, sample: Sample, materials: list[Sample], row: ListRow, *args):
        materials.remove(sample)
        row.destroy()
        self._check_change()

    def _drag_start(self, selection):
        vals = self.collection_tree.item(selection, "values")
        return len(vals) > 1 and vals[1] == 'Element'

    def _on_drop(self, selection, frame, elements):
        vals = self.collection_tree.item(selection, "values")
        sample = self.instance.get_sample(int(vals[0]))
        elements.append(sample)
        self._list_sdd_materials(elements, len(elements) - 1, frame)
        self._check_change()

    def set_style(self, style: ttk.Style):
        super().set_style(style)
        style.configure('Drop.TLabel', foreground="#555555", padx=(0, 5), pady=(0, 5))
        style.configure('Save.TButton', background='#5cb85c', foreground="#ffffff", padx=(0, 5), pady=(0, 5))
        style.configure('Menu.TButton', background='#333333', foreground="#ffffff", padx=(0, 5), pady=(0, 5))
        style.configure('Fav.TButton',  background="#ffffff", relief=tk.SOLID, borderwidth=0, padx=(0, 5), pady=(0, 5))

    def _save(self):
        self.reaction.properties['name'] = self._name_var.get()
        lg = LoadingGIF(self._page_manager.root)
        lg.add_label(self._save_btn)
        lg.start()
        def stop(success):
            lg.stop()
            if success:
                self._origen_data = self._prepare_compare_data()
                self._is_new = False
                self._check_change()
                messagebox.showinfo("Saving Success", "The Reaction was saved.")
            else:
                messagebox.showerror("Saving failed", "The Reaction was not saved.")

        def run():
            try:
                self.reaction.save(True)
                self._page_manager.root.after(1000, stop, True)
            except RequestException as e:
                self._page_manager.root.after(1000, stop, False)

        threading.Thread(target=run).start()

    def _name_change(self, *args):
        self.reaction.properties['name'] = self._name_var.get()
        self._check_change()

    def _check_change(self):
        if self._origen_data == self._prepare_compare_data():
            self._save_btn.state(["disabled"])
            if not self._is_new:
                self._doe_btn.state(["!disabled"])
            else:
                self._doe_btn.state(["disabled"])
        else:
            self._save_btn.state(["!disabled"])
            self._doe_btn.state(["disabled"])

    def _prepare_compare_data(self):
        return "__".join([
            self.reaction.properties['name'],
            "_".join([str(x.id) for x in self.reaction.properties['starting_materials']]),
            "_".join([str(x.id) for x in self.reaction.properties['reactants']]),
            "_".join([str(x.id) for x in self.reaction.properties['products']]),
        ])

    def _style_favorite(self):
        if self.reaction.id in ConfigManager().favorites:
            self._fav_btn.config(image=IconManager().FAVORITE)
        else:
            self._fav_btn.config(image=IconManager().NO_FAVORITE)

    def _toggel_favorites(self):
        if self.reaction.id in ConfigManager().favorites:
            ConfigManager().remove_from_favorites(self.reaction)
        else:
            ConfigManager().add_to_favorites(self.reaction)
        self.update_fav_dropdown()
        self._style_favorite()

    def _go_to_doe(self):
        doe = DoEPage(self.reaction)
        self._page_manager.set_page(doe)

    def _go_to_ana(self):
        ana = AnalysesUploader(self.reaction)
        self._page_manager.set_page(ana)

    def create_new(self, col: Collection):
        self._page_manager.set_page(NewSample(col))